from abc import ABC, abstractmethod
from typing import Optional, List

from resotocore.task.task_description import Job, RunningTask, Workflow
from resotocore.ids import TaskDescriptorId


class TaskHandler(ABC):
    @abstractmethod
    async def list_workflows(self) -> List[Workflow]:
        pass

    @abstractmethod
    async def list_jobs(self) -> List[Job]:
        pass

    @abstractmethod
    async def add_job(self, job: Job) -> None:
        pass

    @abstractmethod
    async def delete_job(self, job_id: str) -> Optional[Job]:
        pass

    @abstractmethod
    async def start_task_by_descriptor_id(self, uid: TaskDescriptorId) -> Optional[RunningTask]:
        pass

    @abstractmethod
    async def running_tasks(self) -> List[RunningTask]:
        pass



from __future__ import annotations

import logging
import uuid
from abc import ABC
from datetime import timedelta
from enum import Enum
from typing import Optional, Any, Sequence, MutableSequence, Callable, Dict, List, Set, Tuple

from attrs import define

from asyncio import Task

from apscheduler.triggers.cron import CronTrigger
from frozendict import frozendict
from jsons import set_deserializer, set_serializer
from transitions import Machine, State, MachineError

from resotocore.ids import TaskId
from resotocore.task.model import Subscriber
from resotocore.message_bus import Event, Action, ActionDone, Message, ActionError
from resotocore.model.typed_model import to_json, from_js, to_js
from resotocore.types import Json
from resotocore.util import first, interleave, empty, exist, identity, utc, utc_str
from resotocore.ids import SubscriberId, TaskDescriptorId


log = logging.getLogger(__name__)


class StepErrorBehaviour(Enum):
    """
    This enumeration defines the behaviour of steps in case of an error:
    - Continue: the response from the actor is ignored and the whole task continues.
    - Stop: the task will be stopped in case of error
    Default is: Continue
    """

    Continue = 1
    Stop = 2


class TaskSurpassBehaviour(Enum):
    """
    This enumeration defines the behaviour of a spawned task where the previous task
    of the same task description is still running.
    - Skip: the new task is not started and dropped.
    - Parallel: the new task is started and runs side by side with the already running instance.
    - Replace: the already running task is stopped and gets replaced by the new one.
    - Wait: wait for the current job to finish and then execute.
            Note: the same task description can only be enqueued once, not multiple times.
    """

    Skip = 1
    Parallel = 2
    Replace = 3
    Wait = 4


# region StepAction: what to do in one step


class StepAction(ABC):
    """
    Base class for an action that should be performed in one step.
    """

    def __eq__(self, other: Any) -> bool:
        return self.__dict__ == other.__dict__ if isinstance(other, StepAction) else False

    @staticmethod
    def from_json(json: Json, _: type = object, **__: object) -> StepAction:
        if "wait_for_message_type" in json:
            return from_js(json, WaitForEvent)
        elif "message_type" in json:
            return from_js(json, PerformAction)
        elif "command" in json:
            return from_js(json, ExecuteCommand)
        elif "event" in json:
            return from_js(json, EmitEvent)
        else:
            raise AttributeError(f"Can not deserialize {json} into StepAction!")


# All actions that need to be restarted from start when the action was interrupted
class RestartAgainStepAction(StepAction):
    pass


@define(order=True, hash=True, frozen=True)
class PerformAction(StepAction):
    # Perform an action by emitting an action message and wait for all subscribers to respond.
    message_type: str


@define(order=True, hash=True, frozen=True)
class EmitEvent(StepAction):
    # Emit this event
    event: Event


@define(order=True, hash=True, frozen=True)
class WaitForEvent(StepAction):
    # Wait for this event to arrive
    wait_for_message_type: str
    filter_data: Optional[Json] = None


@define(order=True, hash=True, frozen=True)
class ExecuteCommand(RestartAgainStepAction):
    # Execute this command in the command interpreter.
    command: str


# endregion

# region StepCommand: resulting commands that get executed. Note that one action can create multiple commands.


class TaskCommand(ABC):
    def __eq__(self, other: Any) -> bool:
        return self.__dict__ == other.__dict__ if isinstance(other, TaskCommand) else False

    @staticmethod
    def from_json(json: Json, _: type = object, **__: object) -> TaskCommand:
        if "message" in json:
            return from_js(json, SendMessage)
        elif "command" in json:
            return from_js(json, ExecuteOnCLI)
        else:
            raise AttributeError(f"Can not deserialize {json} into TaskCommand!")


@define(order=True, hash=True, frozen=True)
class SendMessage(TaskCommand):
    message: Message


@define(order=True, hash=True, frozen=True)
class ExecuteOnCLI(TaskCommand):
    command: str
    # noinspection PyUnresolvedReferences
    env: frozendict  # type: ignore


# endregion

# region Trigger: when a task should be triggered
class Trigger(ABC):
    def __eq__(self, other: object) -> bool:
        return self.__dict__ == other.__dict__ if isinstance(other, Trigger) else False

    @staticmethod
    def from_json(json: Json, _: type = object, **__: object) -> Trigger:
        if "cron_expression" in json:
            return from_js(json, TimeTrigger)
        elif "message_type" in json:
            return from_js(json, EventTrigger)
        else:
            raise AttributeError(f"Can not deserialize {json} into StepAction!")


@define(order=True, hash=True, frozen=True)
class EventTrigger(Trigger):
    message_type: str
    filter_data: Optional[Json] = None


@define(order=True, hash=True, frozen=True)
class TimeTrigger(Trigger):
    cron_expression: str

    def __attrs_post_init__(self) -> None:
        # make sure the time trigger is valid
        CronTrigger.from_crontab(self.cron_expression)


# endregion


class Step:
    """
    Immutable description of a step inside a task.
    """

    def __init__(
        self,
        name: str,
        action: StepAction,
        timeout: Optional[timedelta] = None,
        on_error: StepErrorBehaviour = StepErrorBehaviour.Continue,
    ):
        self.name = name
        self.action = action
        self.timeout = timeout if timeout else timedelta(minutes=5)
        self.on_error = on_error

    def __eq__(self, other: object) -> bool:
        return self.__dict__ == other.__dict__ if isinstance(other, Step) else False


class TaskDescription(ABC):
    def __init__(
        self,
        uid: TaskDescriptorId,
        name: str,
        steps: Sequence[Step],
        triggers: Sequence[Trigger],
        on_surpass: TaskSurpassBehaviour,
        environment: Optional[Dict[str, str]],
        mutable: bool,
    ):
        self.id = uid
        self.name = name
        self.steps = steps
        self.triggers = triggers
        self.on_surpass = on_surpass
        self.environment = environment if environment else {}
        self.mutable = mutable

    def step_by_name(self, name: str) -> Optional[Step]:
        return first(lambda x: x.name == name, self.steps)

    def __eq__(self, other: object) -> bool:
        return self.__dict__ == other.__dict__ if isinstance(other, TaskDescription) else False


class Job(TaskDescription):
    def __init__(
        self,
        uid: TaskDescriptorId,
        command: ExecuteCommand,
        timeout: timedelta,
        trigger: Optional[Trigger] = None,
        wait: Optional[Tuple[EventTrigger, timedelta]] = None,
        environment: Optional[Dict[str, str]] = None,
        mutable: bool = True,
        active: bool = True,
    ):
        steps: List[Step] = []
        if wait:
            wait_trigger, wait_timeout = wait
            action = WaitForEvent(wait_trigger.message_type, wait_trigger.filter_data)
            steps.append(Step("wait", action, wait_timeout, StepErrorBehaviour.Stop))
        steps.append(Step("execute", command, timeout, StepErrorBehaviour.Stop))
        start_trigger = [trigger] if active and trigger else []
        super().__init__(uid, uid, steps, start_trigger, TaskSurpassBehaviour.Wait, environment, mutable)
        self.command = command
        self.timeout = timeout
        self.trigger = trigger
        self.wait = wait
        self.active = active

    @staticmethod
    def to_json(o: Job, **_: object) -> Json:
        wait = {"wait_trigger": to_js(o.wait[0]), "wait_timeout": to_json(o.wait[1])} if o.wait else {}
        env = {"environment": o.environment} if o.environment else {}
        return {
            "id": o.id,
            "name": o.name,
            "command": to_js(o.command),
            "trigger": to_js(o.trigger),
            "timeout": to_json(o.timeout),
            "active": o.active,
            **env,
            **wait,
        }

    @staticmethod
    def from_json(json: Json, _: type = object, **__: object) -> Job:
        maybe_wait = (
            (from_js(json["wait_trigger"], EventTrigger), from_js(json["wait_timeout"], timedelta))
            if "wait_trigger" in json
            else None
        )
        trigger = json.get("trigger")
        return Job(
            json["id"],
            from_js(json["command"], ExecuteCommand),
            from_js(json["timeout"], timedelta),
            from_js(trigger, Trigger) if trigger is not None else None,
            maybe_wait,
            json.get("environment"),
            active=json.get("active", True),  # backward compatibility: in case the prop is missing
        )


class Workflow(TaskDescription):
    """
    Immutable description of a complete workflow.
    """

    def __init__(
        self,
        uid: TaskDescriptorId,
        name: str,
        steps: Sequence[Step],
        triggers: Sequence[Trigger],
        on_surpass: TaskSurpassBehaviour = TaskSurpassBehaviour.Skip,
        environment: Optional[Dict[str, str]] = None,
    ) -> None:
        super().__init__(uid, name, steps, triggers, on_surpass, environment, mutable=False)
        self._triggers = triggers
        self._on_surpass = on_surpass

    @staticmethod
    def to_json(o: Workflow, **_: object) -> Json:
        env = {"environment": o.environment} if o.environment else {}
        return {
            "id": o.id,
            "name": o.name,
            "steps": to_json(o.steps),
            "triggers": to_json(o.triggers),
            "on_surpass": to_js(o.on_surpass),
            **env,
        }

    @staticmethod
    def from_json(json: Json, _: type = object, **__: object) -> Workflow:
        return Workflow(
            json["id"],
            json["name"],
            from_js(json["steps"], List[Step]),
            from_js(json["triggers"], List[Trigger]),
            from_js(json["on_surpass"], TaskSurpassBehaviour),
        )


class StepState(State):  # type: ignore
    """
    Base class for all states in a task.
    There is always a related step definition inside a related task definition.
    """

    def __init__(self, step: Step, instance: RunningTask):
        super().__init__(step.name, "begin_step")
        self.step = step
        self.instance = instance
        self.timed_out = False

    def current_step_done(self) -> bool:
        """
        Override this method in deriving classes to define, when this state is done.
        :return: True when this state is done otherwise false.
        """
        return not self.timed_out

    def commands_to_execute(self) -> Sequence[TaskCommand]:
        """
        Override this method in deriving classes to define actions when this state is entered.
        :return: all commands to execute when this state is entered.
        """
        return []

    # noinspection PyUnusedLocal
    def handle_command_results(self, results: Dict[TaskCommand, Any]) -> None:
        """
        Override this method in deriving classes to define behaviour based on command results.
        :param results: the results of emitted task commands.
        :return: the list of new task commands.
        """

    def timeout(self) -> timedelta:
        """
        Define the timeout of this step.
        Defaults to the configured step timeout, but can be overridden in subsequent classes.
        :return: the timeout of this step.
        """
        return self.step.timeout

    def handle_event(self, event: Event) -> bool:
        """
        Return true if the internal state of the fsm has changed by this event.
        Note: all events received by the system are provided to this state.
              The state has to decide if this event is meaningful or not.
        """
        return False

    def check_timeout(self) -> bool:
        """
        Return true if the internal state of the fsm has changed by this event.
        This method is called periodically by the cleaner task.
        """
        if (self.instance.step_started_at + self.timeout()) < utc():
            self.timed_out = True
            return True
        return False

    @staticmethod
    def from_step(step: Step, instance: RunningTask) -> StepState:
        """
        Create the related state based on the given step and task description.
        """
        if isinstance(step.action, PerformAction):
            return PerformActionState(step.action, step, instance)
        elif isinstance(step.action, EmitEvent):
            return EmitEventState(step.action, step, instance)
        elif isinstance(step.action, WaitForEvent):
            return WaitForEventState(step.action, step, instance)
        elif isinstance(step.action, ExecuteCommand):
            return ExecuteCommandState(step.action, step, instance)
        else:
            raise AttributeError(f"No mapping for {type(step.action).__name__}")

    def step_started(self) -> None:
        """
        This method is called when the fsm enters this state.
        Override in subsequent classes, if action is required in such a scenario.
        """

    # noinspection PyMethodMayBeStatic
    def export_state(self) -> Json:
        """
        This method is called when the state of the task needs to be persisted.
        Since each state in the FSM can have it's own schema, we export a generic json blob here,
        that has to be interpreted during import_state.
        :return: json representation of this state. empty by default.
        """
        return {}

    def import_state(self, js: Json) -> None:
        """
        This method is called when the execution of this task has been interrupted by a restart.
        The last known state is persisted to some durable storage and imported in the startup phase.

        :param js: the same json that was exported with export_state()
        """


class PerformActionState(StepState):
    """
    This state emits an action when started and then waits for all actors to respond with a done message.
    State is done, when all subscribers with expected answer send a done message.
    """

    def __init__(self, perform: PerformAction, step: Step, instance: RunningTask):
        super().__init__(step, instance)
        self.perform = perform
        self.wait_for: List[Subscriber] = self.instance.subscribers_by_event().get(perform.message_type, [])

    def current_step_done(self) -> bool:
        """
        This state is done, when we received an ack or an error from every subscriber.
        The step behavior defines how to deal in case of an error.
        """
        msg_type = self.perform.message_type
        in_step: Set[SubscriberId] = {
            x.subscriber_id
            for x in self.instance.received_messages
            if isinstance(x, (ActionDone, ActionError)) and x.step_name == self.step.name
        }
        subscriber = self.wait_for
        missing = {x.id for x in subscriber if x[msg_type].wait_for_completion} - in_step
        return self.timed_out or (not self.instance.is_error and empty(missing))

    def timeout(self) -> timedelta:
        """
        The timeout is extended to the longest timeout of all subscribers falling back to the step timeout.
        """
        msg_type = self.perform.message_type
        max_timeout = self.step.timeout
        for subscriber in self.instance.subscribers_by_event().get(msg_type, []):
            subscription = subscriber.subscriptions[msg_type]
            to = subscription.timeout
            # only extend the timeout, when the subscriber is blocking and has a longer timeout
            max_timeout = max_timeout if not subscription.wait_for_completion or max_timeout > to else to
        return max_timeout

    def commands_to_execute(self) -> Sequence[TaskCommand]:
        """
        When the state is entered, emit the action message and inform all actors.
        """
        return [SendMessage(Action(self.perform.message_type, self.instance.id, self.step.name))]

    def step_started(self) -> None:
        # refresh the list of subscribers when the step has started
        self.wait_for = self.instance.subscribers_by_event().get(self.perform.message_type, [])

    def export_state(self) -> Json:
        return {"wait_for": [a.id for a in self.wait_for]}

    def import_state(self, js: Json) -> None:
        existing = {s.id: s for s in self.instance.subscribers_by_event().get(self.perform.message_type, [])}
        wait_for = js.get("wait_for", [])
        # filter all existing subscriber from the list of subscribers to wait_for
        self.wait_for = list(filter(identity, (existing.get(sid) for sid in wait_for)))  # type: ignore


class WaitForEventState(StepState):
    def __init__(self, perform: WaitForEvent, step: Step, instance: RunningTask):
        super().__init__(step, instance)
        self.perform = perform

    def current_step_done(self) -> bool:
        """
        This step is done, when the event it is waiting for has arrived.
        """
        return self.timed_out or exist(
            lambda x: isinstance(x, Event) and x.message_type == self.perform.wait_for_message_type,
            self.instance.received_messages,
        )

    def handle_event(self, event: Event) -> bool:
        """
        Check if the provided event is the one this step is waiting for.
        The event has to have the same message_type and the provided filter has to apply.
        """

        def filter_applies() -> bool:
            comp = self.perform.filter_data
            if comp:
                return {key: event.data.get(key) for key in comp} == comp
            else:
                return True

        if event.message_type == self.perform.wait_for_message_type and filter_applies():
            self.instance.received_messages.append(event)
            return True
        return False


class ExecuteCommandState(StepState):
    def __init__(self, execute: ExecuteCommand, step: Step, instance: RunningTask):
        super().__init__(step, instance)
        self.execute = execute
        self.execution_done = False

    def commands_to_execute(self) -> Sequence[TaskCommand]:
        # override now: always use the time when the task has been triggered
        env = frozendict({"now": utc_str(self.instance.task_started_at)})
        return [ExecuteOnCLI(self.execute.command, env)]

    def handle_command_results(self, results: Dict[TaskCommand, Any]) -> None:
        found = first(lambda r: isinstance(r, ExecuteOnCLI) and r.command == self.execute.command, results.keys())
        if found:
            result = results[found]
            if isinstance(result, Exception):
                log.warning(f"Command {self.execute.command} failed with error: {result}")
            else:
                log.info(f"Result of command {self.execute.command} is {result}")
            self.execution_done = True

    def current_step_done(self) -> bool:
        return self.execution_done


class EmitEventState(StepState):
    def __init__(self, emit: EmitEvent, step: Step, instance: RunningTask):
        super().__init__(step, instance)
        self.emit = emit

    def commands_to_execute(self) -> Sequence[TaskCommand]:
        return [SendMessage(self.emit.event)]


class StartState(StepState):
    def __init__(self, instance: RunningTask):
        self.event = Event("task_start")
        super().__init__(Step("task_start", EmitEvent(self.event)), instance)

    def commands_to_execute(self) -> Sequence[TaskCommand]:
        return [SendMessage(self.event)]


class EndState(StepState):
    """
    This state marks the end of the task.
    """

    def __init__(self, instance: RunningTask):
        self.event = Event("task_end")
        super().__init__(Step("task_end", EmitEvent(self.event)), instance)

    def is_error(self) -> bool:
        return self.instance.is_error

    def current_step_done(self) -> bool:
        return False

    def commands_to_execute(self) -> Sequence[TaskCommand]:
        return [SendMessage(self.event)]


class RunningTask:
    @staticmethod
    def empty(
        descriptor: TaskDescription, subscriber_by_event: Callable[[], Dict[str, List[Subscriber]]]
    ) -> Tuple[RunningTask, Sequence[TaskCommand]]:
        assert len(descriptor.steps) > 0, "TaskDescription needs at least one step!"
        uid = TaskId(str(uuid.uuid1()))
        task = RunningTask(uid, descriptor, subscriber_by_event)
        messages = [SendMessage(Event("task_started", data={"task": descriptor.name})), *task.move_to_next_state()]
        return task, messages

    def __init__(
        self, uid: TaskId, descriptor: TaskDescription, subscribers_by_event: Callable[[], Dict[str, List[Subscriber]]]
    ):
        self.id = uid
        self.is_error = False
        self.descriptor = descriptor
        self.received_messages: MutableSequence[Message] = []
        self.subscribers_by_event = subscribers_by_event
        self.task_started_at = utc()
        self.step_started_at = self.task_started_at
        self.update_task: Optional[Task[None]] = None
        self.descriptor_alive = True

        steps = [StepState.from_step(step, self) for step in descriptor.steps]
        start = StartState(self)
        end = EndState(self)
        states: List[StepState] = [start, *steps, end]
        self.machine = Machine(self, states, start, auto_transitions=False, queued=True)

        for current_state, next_state in interleave(states):
            self.machine.add_transition(
                "_next_state", current_state.name, next_state.name, [current_state.current_step_done]
            )
            self.machine.add_transition("_to_err", current_state.name, end.name, [end.is_error])

    def move_to_next_state(self) -> Sequence[TaskCommand]:
        def next_state() -> bool:
            try:
                # this method is defined dynamically by transitions
                last_state = self.current_state
                result: bool = self._next_state()  # type: ignore # pylint: disable=no-member
                # safe guard: if state transition does not change the state
                if result and self.current_state is last_state:
                    return False
                return result
            except MachineError:
                return False

        resulting_commands: List[TaskCommand] = []
        while next_state():
            resulting_commands.extend(self.current_state.commands_to_execute())
        return resulting_commands

    @property
    def current_state(self) -> StepState:
        return self.machine.get_state(self.state)  # type: ignore # pylint: disable=no-member

    @property
    def current_step(self) -> Step:
        return self.current_state.step

    @property
    def is_active(self) -> bool:
        return not isinstance(self.current_state, EndState)

    def handle_event(self, event: Event) -> Tuple[bool, Sequence[TaskCommand]]:
        if self.current_state.handle_event(event):
            return True, self.move_to_next_state()
        else:
            return False, []

    def handle_done(self, done: ActionDone) -> Sequence[TaskCommand]:
        self.received_messages.append(done)
        return self.move_to_next_state()

    def handle_error(self, error: ActionError) -> Sequence[TaskCommand]:
        """
        An action could not be performed - the subscriber returned an error.
        Such a message only makes sense in a PerformAction step.
        Whether or not this event leads to a state change is decided by the current state.
        Whether or not this leads to the end of this task is decided by the current step error behaviour.
        """
        if not isinstance(self.current_step.action, PerformAction):
            log.info(
                f"Received action error {error} but the current step is "
                f"{type(self.current_step.action).__name__}. Ignore."
            )
            return []
        elif self.is_active and self.current_step.on_error == StepErrorBehaviour.Continue:
            self.received_messages.append(error)
            return self.move_to_next_state()
        else:
            log.info(
                f"Task: {error.task_id}: Subscriber {error.subscriber_id} could not handle action: "
                f"{error.message_type} because: {error.error}. Stop this task."
            )
            self.end()
            return []

    def handle_command_results(self, results: Dict[TaskCommand, Any]) -> Sequence[TaskCommand]:
        self.current_state.handle_command_results(results)
        return self.move_to_next_state()

    def end(self) -> None:
        """
        If this method is called, the task is marked as failed and moves to the end state.
        Use this method to abort a task.
        """
        if not isinstance(self.current_state, EndState):
            self.is_error = True
            self._to_err()  # type: ignore # pylint: disable=no-member

    def ack_for(self, message_type: str, subscriber: Subscriber) -> Optional[Message]:
        """
        Return the ack received ack for the given message_type of the given subscriber or None.
        """

        def relevant_ack(message: Message) -> bool:
            return (
                isinstance(message, (ActionDone, ActionError))
                and message.message_type == message_type
                and message.subscriber_id == subscriber.id
            )

        return first(relevant_ack, self.received_messages)

    def pending_action_for(self, subscriber: Subscriber) -> Optional[Action]:
        """
        In case this task is waiting for an action result from the given subscriber,
        the relevant action is returned.
        """
        state = self.current_state
        if isinstance(state, PerformActionState):
            message_type = state.perform.message_type
            subscriptions = state.wait_for
            if subscriber in subscriptions and self.ack_for(message_type, subscriber) is None:
                return Action(message_type, self.id, state.step.name)
        return None

    def begin_step(self) -> None:
        log.info(f"Task {self.id}: begin step is: {self.current_step.name}")
        # update the step started time, whenever a new state is entered
        self.step_started_at = utc()
        self.current_state.step_started()


set_deserializer(StepAction.from_json, StepAction, high_prio=False)
set_deserializer(Trigger.from_json, Trigger, high_prio=False)
set_deserializer(TaskCommand.from_json, TaskCommand, high_prio=False)
set_deserializer(Job.from_json, Job)
set_serializer(Job.to_json, Job)
set_deserializer(Workflow.from_json, Workflow)
set_serializer(Workflow.to_json, Workflow)



from __future__ import annotations

import inspect
import json
from abc import ABC, abstractmethod
from asyncio import Queue, Task, iscoroutine
from attrs import define, field
from enum import Enum
from operator import attrgetter
from textwrap import dedent
from typing import Optional, List, Any, Dict, Tuple, Callable, Union, Awaitable, Type, cast, Set

from aiohttp import ClientSession, TCPConnector
from aiostream import stream
from aiostream.core import Stream
from rich.jupyter import JupyterMixin
from parsy import test_char, string

from resotolib.parse_util import l_curly_dp, r_curly_dp
from resotocore.analytics import AnalyticsEventSender
from resotocore.cli import JsGen, T, Sink
from resotocore.config import ConfigHandler
from resotocore.core_config import CoreConfig, AliasTemplateConfig, AliasTemplateParameterConfig
from resotocore.db.db_access import DbAccess
from resotocore.error import CLIParseError
from resotocore.console_renderer import ConsoleRenderer, ConsoleColorSystem
from resotocore.message_bus import MessageBus
from resotocore.model.model_handler import ModelHandler
from resotocore.query.model import Query, variable_to_absolute, PathRoot
from resotocore.query.template_expander import TemplateExpander, render_template
from resotocore.task import TaskHandler
from resotocore.types import Json, JsonElement
from resotocore.util import AccessJson
from resotocore.web.certificate_handler import CertificateHandler
from resotocore.worker_task_queue import WorkerTaskQueue


class MediaType(Enum):
    Json = 1
    FilePath = 2

    @property
    def json(self) -> bool:
        return self == MediaType.Json

    @property
    def file_path(self) -> bool:
        return self == MediaType.FilePath

    def __repr__(self) -> str:
        return "application/json" if self == MediaType.Json else "application/octet-stream"


no_closing_p = test_char(lambda x: x != "}", "No closing bracket").at_least(1).concat()
no_bracket_p = test_char(lambda x: x not in ("{", "}"), "No opening bracket").at_least(1).concat()
double_curly_open_dp = string("{{")
double_curly_close_dp = string("}}")
l_or_r_curly_dp = string("{") | string("}")


@define(frozen=True)
class CLIContext:
    env: Dict[str, str] = field(factory=dict)
    uploaded_files: Dict[str, str] = field(factory=dict)  # id -> path
    query: Optional[Query] = None
    query_options: Dict[str, Any] = field(factory=dict)
    console_renderer: Optional[ConsoleRenderer] = None
    source: Optional[str] = None  # who is calling

    def variable_in_section(self, variable: str) -> str:
        # if there is no query, always assume the root section
        section = self.env.get("section") if self.query else PathRoot
        return variable_to_absolute(section, variable)

    def render_console(self, element: Union[str, JupyterMixin]) -> str:
        if self.console_renderer:
            return self.console_renderer.render(element)
        elif isinstance(element, JupyterMixin):
            return str(element)
        else:
            return element

    def supports_color(self) -> bool:
        return (
            self.console_renderer is not None
            and self.console_renderer.color_system is not None
            and self.console_renderer.color_system != ConsoleColorSystem.monochrome
        )

    def formatter(self, format_string: str) -> Callable[[Json], str]:
        return self.formatter_with_variables(format_string, False)[0]

    def formatter_with_variables(
        self, format_string: str, collect_variables: bool = True
    ) -> Tuple[Callable[[Json], str], Optional[Set[str]]]:
        """
        A renderer can be used to string format objects based on a provided format string.
        """

        variables: Optional[Set[str]] = set() if collect_variables else None

        def format_variable(name: str) -> str:
            assert "__" not in name, "No dunder attributes allowed"
            in_section = self.variable_in_section(name)
            if collect_variables:
                variables.add(in_section)  # type: ignore
            return "{" + in_section + "}"

        def render_simple_property(prop: Any) -> str:
            return json.dumps(prop) if isinstance(prop, bool) else str(prop)

        variable = (l_curly_dp >> no_closing_p << r_curly_dp).map(format_variable)
        token = double_curly_open_dp | double_curly_close_dp | no_bracket_p | variable | l_or_r_curly_dp
        format_string_parser = token.many().concat()
        formatter: str = format_string_parser.parse(format_string)

        def format_object(obj: Any) -> str:
            return formatter.format_map(AccessJson.wrap(obj, "null", render_simple_property))

        return format_object, variables


EmptyContext = CLIContext()


class CLIEngine(ABC):
    @abstractmethod
    async def evaluate_cli_command(
        self, cli_input: str, context: CLIContext = EmptyContext, replace_place_holder: bool = True
    ) -> List[ParsedCommandLine]:
        pass


class CLIDependencies:
    def __init__(self, **deps: Any) -> None:
        self.lookup: Dict[str, Any] = deps

    def extend(self, **deps: Any) -> CLIDependencies:
        self.lookup = {**self.lookup, **deps}
        return self

    @property
    def config(self) -> CoreConfig:
        return self.lookup["config"]  # type: ignore

    @property
    def message_bus(self) -> MessageBus:
        return self.lookup["message_bus"]  # type:ignore

    @property
    def event_sender(self) -> AnalyticsEventSender:
        return self.lookup["event_sender"]  # type:ignore

    @property
    def db_access(self) -> DbAccess:
        return self.lookup["db_access"]  # type:ignore

    @property
    def model_handler(self) -> ModelHandler:
        return self.lookup["model_handler"]  # type:ignore

    @property
    def task_handler(self) -> TaskHandler:
        return self.lookup["task_handler"]  # type:ignore

    @property
    def worker_task_queue(self) -> WorkerTaskQueue:
        return self.lookup["worker_task_queue"]  # type:ignore

    @property
    def template_expander(self) -> TemplateExpander:
        return self.lookup["template_expander"]  # type:ignore

    @property
    def forked_tasks(self) -> Queue[Tuple[Task[JsonElement], str]]:
        return self.lookup["forked_tasks"]  # type:ignore

    @property
    def cli(self) -> CLIEngine:
        return self.lookup["cli"]  # type:ignore

    @property
    def config_handler(self) -> ConfigHandler:
        return self.lookup["config_handler"]  # type:ignore

    @property
    def cert_handler(self) -> CertificateHandler:
        return self.lookup["cert_handler"]  # type:ignore

    @property
    def http_session(self) -> ClientSession:
        session: Optional[ClientSession] = self.lookup.get("http_session")
        if not session:
            connector = TCPConnector(limit=0, ssl=False, ttl_dns_cache=300)
            session = ClientSession(connector=connector)
            self.lookup["http_session"] = session
        return session

    async def stop(self) -> None:
        if "http_session" in self.lookup:
            await self.http_session.close()


@define
class CLICommandRequirement:
    name: str


@define
class CLIFileRequirement(CLICommandRequirement):
    path: str  # local client path


class CLIAction(ABC):
    def __init__(
        self, produces: MediaType, requires: Optional[List[CLICommandRequirement]], envelope: Optional[Dict[str, str]]
    ) -> None:
        self.produces = produces
        self.required = requires or []
        self.envelope: Dict[str, str] = envelope or {}

    @staticmethod
    def make_stream(in_stream: JsGen) -> Stream:
        return in_stream if isinstance(in_stream, Stream) else stream.iterate(in_stream)


class CLISource(CLIAction):
    def __init__(
        self,
        fn: Callable[[], Union[Tuple[Optional[int], JsGen], Awaitable[Tuple[Optional[int], JsGen]]]],
        produces: MediaType = MediaType.Json,
        requires: Optional[List[CLICommandRequirement]] = None,
        envelope: Optional[Dict[str, str]] = None,
    ) -> None:
        super().__init__(produces, requires, envelope)
        self._fn = fn

    async def source(self) -> Tuple[Optional[int], Stream]:
        res = self._fn()
        count, gen = await res if iscoroutine(res) else res
        return count, self.make_stream(await gen if iscoroutine(gen) else gen)

    @staticmethod
    def with_count(
        fn: Callable[[], Union[JsGen, Awaitable[JsGen]]],
        count: Optional[int],
        produces: MediaType = MediaType.Json,
        requires: Optional[List[CLICommandRequirement]] = None,
        envelope: Optional[Dict[str, str]] = None,
    ) -> CLISource:
        async def combine() -> Tuple[Optional[int], JsGen]:
            res = fn()
            gen = await res if iscoroutine(res) else res
            return count, gen

        return CLISource(combine, produces, requires, envelope)

    @staticmethod
    def single(
        fn: Callable[[], Union[JsGen, Awaitable[JsGen]]],
        produces: MediaType = MediaType.Json,
        requires: Optional[List[CLICommandRequirement]] = None,
        envelope: Optional[Dict[str, str]] = None,
    ) -> CLISource:
        return CLISource.with_count(fn, 1, produces, requires, envelope)

    @staticmethod
    def empty() -> CLISource:
        return CLISource.with_count(stream.empty, 0)


class CLIFlow(CLIAction):
    def __init__(
        self,
        fn: Callable[[JsGen], Union[JsGen, Awaitable[JsGen]]],
        produces: MediaType = MediaType.Json,
        requires: Optional[List[CLICommandRequirement]] = None,
        envelope: Optional[Dict[str, str]] = None,
    ) -> None:
        super().__init__(produces, requires, envelope)
        self._fn = fn

    async def flow(self, in_stream: JsGen) -> Stream:
        gen = self._fn(self.make_stream(in_stream))
        return self.make_stream(await gen if iscoroutine(gen) else gen)


@define
class ArgInfo:
    # If the argument has a name. It is quite common that arguments do not have a name
    # but are expected at some position.
    # Example: `count <kind>`: kind is the argument at position 1 without a name
    name: Optional[str] = None
    # Defines if this argument expects a value.
    # Some arguments are only flags, while others expect a value.
    # Example: `--compress` is a flag without value
    # Example: `--count <kind>` is an argument with value
    expects_value: bool = False
    # If the value has to be picked from a list of values (enumeration).
    # Example: `--format svg|png|jpg`
    possible_values: List[str] = field(factory=list)
    # If this argument is allowed to be specified multiple times
    can_occur_multiple_times: bool = False
    # Give a type hint for the argument value.
    # Allowed values are:
    # - `file`: the argument expects a file path
    # - `kind`: the argument expects a kind in the model
    # - `property`: the argument expects a property in the model
    # - `command`: the argument expects a command on the cli
    # - `event`: the event handled or emitted by the task handler
    # - `search`: the argument expects a search string
    value_hint: Optional[str] = None
    # Help text of the argument option.
    help_text: Optional[str] = None
    # If multiple options share the same group, only one of them can be selected.
    # Use groups if you have multiple options, where only one is allowed to be selected.s
    option_group: Optional[str] = None


# mypy does not support recursive type aliases: define 3 levels as maximum here
ArgsInfo = Union[
    Dict[str, Union[Dict[str, Union[Dict[str, Union[Any, List[ArgInfo]]], List[ArgInfo]]], List[ArgInfo]]],
    List[ArgInfo],
]


class CLICommand(ABC):
    """
    The CLIPart is the base for all participants of the cli execution.
    Source: generates a stream of objects
    Flow: transforms the elements in a stream of objects
    Sink: takes a stream of objects and creates a result
    """

    def __init__(self, dependencies: CLIDependencies, allowed_in_source_position: bool = False) -> None:
        self.dependencies = dependencies
        self.allowed_in_source_position = allowed_in_source_position

    @property
    @abstractmethod
    def name(self) -> str:
        pass

    def help(self) -> str:
        # if not defined in subclass, fallback to inline doc
        doc = inspect.getdoc(type(self))
        return doc if doc else f"{self.name}: no help available."

    def rendered_help(self, ctx: CLIContext) -> str:
        text = f"\n**{self.name}: {self.info()}**\n\n{self.help()}"
        return ctx.render_console(text)

    @abstractmethod
    def args_info(self) -> ArgsInfo:
        pass

    @abstractmethod
    def info(self) -> str:
        pass

    @abstractmethod
    def parse(self, arg: Optional[str] = None, ctx: CLIContext = EmptyContext, **kwargs: Any) -> CLIAction:
        pass


@define(order=True, hash=True, frozen=True)
class AliasTemplateParameter:
    name: str
    description: str
    default: Optional[JsonElement] = None

    def example_value(self) -> JsonElement:
        return self.default if self.default else f"test_{self.name}"


# pylint: disable=not-an-iterable
@define(order=True, hash=True, frozen=True)
class AliasTemplate:
    name: str
    info: str
    template: str
    parameters: List[AliasTemplateParameter] = field(factory=list)

    def render(self, props: Json) -> str:
        return render_template(self.template, props)

    def help(self) -> str:
        args = ", ".join(f"{arg.name}=<value>" for arg in self.parameters)

        def param_info(p: AliasTemplateParameter) -> str:
            default = f" [default: {p.default}]" if p.default else ""
            return f"- `{p.name}`{default}: {p.description}"

        indent = "                "
        arg_info = f"\n{indent}".join(param_info(arg) for arg in sorted(self.parameters, key=attrgetter("name")))
        minimal = ", ".join(f'{p.name}="{p.example_value()}"' for p in self.parameters if p.default is None)
        return dedent(
            f"""
                {self.name}: {self.info}
                ```shell
                {self.name} {args}
                ```
                ## Parameters
                {arg_info}

                ## Template
                ```shell
                > {self.template}
                ```

                ## Example
                ```shell
                # Executing this alias template
                > {self.name} {minimal}
                # Will expand to this command
                > {self.render({p.name: p.example_value() for p in self.parameters})}
                ```
                """
        )

    def rendered_help(self, ctx: CLIContext) -> str:
        return ctx.render_console(self.help())

    @staticmethod
    def from_config(cfg: AliasTemplateConfig) -> AliasTemplate:
        def arg(p: AliasTemplateParameterConfig) -> AliasTemplateParameter:
            return AliasTemplateParameter(p.name, p.description, p.default)

        return AliasTemplate(cfg.name, cfg.info, cfg.template, [arg(a) for a in cfg.parameters])


class InternalPart(ABC):
    """
    Internal parts can be executed but are not shown via help.
    They usually get injected by the CLI Interpreter to ease usability.
    """


class OutputTransformer(ABC):
    """
    Mark all commands that transform the output stream (formatting).
    """


class PreserveOutputFormat(ABC):
    """
    Mark all commands where the output should not be flattened to default line output.
    """


class NoTerminalOutput(ABC):
    """
    Mark all commands where the output should not contain any terminal escape codes.
    """


@define
class ParsedCommand:
    cmd: str
    args: Optional[str] = None


@define
class ParsedCommands:
    commands: List[ParsedCommand]
    env: Json = field(factory=dict)


@define
class ExecutableCommand:
    name: str  # the name of the command or alias
    command: CLICommand
    arg: Optional[str]
    action: CLIAction


@define
class ParsedCommandLine:
    """
    The parsed command line holds:
    - ctx: the resulting environment coming from the parsed environment + the provided environment
    - commands: all commands this command is defined from
    - generator: this generator can be used in order to execute the command line
    """

    ctx: CLIContext
    parsed_commands: ParsedCommands
    executable_commands: List[ExecutableCommand]
    unmet_requirements: List[CLICommandRequirement]
    envelope: Dict[str, str]

    def __attrs_post_init__(self) -> None:
        def expect_action(cmd: ExecutableCommand, expected: Type[T]) -> T:
            action = cmd.action
            if isinstance(action, expected):
                return action
            else:
                message = "must be the first command" if issubclass(type(action), CLISource) else "no source data given"
                raise CLIParseError(f"Command >{cmd.command.name}< can not be used in this position: {message}")

        if self.executable_commands:
            expect_action(self.executable_commands[0], CLISource)
            for command in self.executable_commands[1:]:
                expect_action(command, CLIFlow)

    async def to_sink(self, sink: Sink[T]) -> T:
        _, generator = await self.execute()
        return await sink(generator)

    @property
    def commands(self) -> List[CLICommand]:
        return [part.command for part in self.executable_commands]

    @property
    def produces(self) -> MediaType:
        # the last command in the chain defines the resulting media type
        return self.executable_commands[-1].action.produces if self.executable_commands else MediaType.Json

    async def execute(self) -> Tuple[Optional[int], Stream]:
        if self.executable_commands:
            source_action = cast(CLISource, self.executable_commands[0].action)
            count, flow = await source_action.source()
            for command in self.executable_commands[1:]:
                flow_action = cast(CLIFlow, command.action)
                flow = await flow_action.flow(flow)
            return count, flow
        else:
            return 0, stream.empty()



from __future__ import annotations

import json
import re
import sys
from abc import ABC, abstractmethod
from attrs import define
from datetime import datetime, timezone, date
from json import JSONDecodeError
from typing import Union, Any, Optional, Callable, Type, Sequence, Dict, List, Set, cast, Tuple, Iterable

import yaml
from dateutil.parser import parse
from jsons import set_deserializer, set_serializer
from networkx import MultiDiGraph
from parsy import regex, string, Parser

from resotolib.durations import duration_parser, DurationRe
from resotolib.parse_util import make_parser, variable_dp_backtick, dot_dp
from resotocore.model.transform_kind_convert import converters
from resotocore.model.typed_model import from_js
from resotocore.types import Json, JsonElement, ValidationResult, ValidationFn, EdgeType
from resotocore.util import if_set, utc, duration, first
from resotocore.compat import remove_suffix


def check_type_fn(t: type, type_name: str) -> ValidationFn:
    def check_type(x: Any) -> ValidationResult:
        if isinstance(x, t):
            return None
        else:
            raise AttributeError(f"Expected type {type_name} but got {type(x).__name__}")

    return check_type


def check_fn(x: Optional[Any], func: Callable[[Any, Any], Optional[Any]], message: str) -> Optional[ValidationFn]:
    def check_single(value: Any) -> ValidationResult:
        if func(x, value):
            return None
        else:
            raise AttributeError(f">{value}< {message}")

    return None if x is None else check_single


def validate_fn(*fns: Optional[ValidationFn]) -> ValidationFn:
    defined = list(filter(lambda x: x is not None, fns))

    def always_valid(_: Any) -> ValidationResult:
        return None

    def check_defined(value: Any) -> ValidationResult:
        for fn in defined:
            res = fn(value)  # type: ignore
            if res is not None:
                return res
        return None

    return check_defined if defined else always_valid


@define(order=True, hash=True, frozen=True)
class SyntheticProperty:
    path: List[str]
    """
    A synthetic property does not exist in the underlying data model.
    It is defined by a function on an existing other property.
    Example: age is a duration defined on ctime which is a datetime.
             the function is age=now-ctime.
    """


@define(order=True, hash=True, frozen=True)
class Property:
    name: str
    kind: str
    required: bool = False
    synthetic: Optional[SyntheticProperty] = None
    description: Optional[str] = None

    def __attrs_post_init__(self) -> None:
        assert self.synthetic is None or not self.required, "Synthetic properties can not be required!"

    def resolve(self, model: Dict[str, Kind]) -> Kind:
        return Property.parse_kind(self.kind, model)

    @staticmethod
    def parse_kind(name: str, model: Dict[str, Kind]) -> Kind:
        def kind_by_name(kind_name: str) -> Kind:
            if kind_name not in model:
                raise AttributeError(f"Property kind is not known: {kind_name}. Have you registered it?")
            return model[kind_name]

        simple_kind_parser = regex("[A-Za-z][A-Za-z0-9_.]*").map(kind_by_name)
        bracket_parser = string("[]")
        dict_string_parser = string("dictionary[")
        comma_parser = regex("\\s*,\\s*")
        bracket_r = string("]")

        @make_parser
        def array_parser() -> Parser:
            inner = yield dictionary_parser | simple_kind_parser
            brackets = yield bracket_parser.times(1, float("inf"))
            return ArrayKind.mk_array(inner, len(brackets))

        @make_parser
        def dictionary_parser() -> Parser:
            yield dict_string_parser
            key_kind = cast(Kind, (yield simple_kind_parser))
            yield comma_parser
            value_kind = yield array_parser | dictionary_parser | simple_kind_parser
            yield bracket_r
            return DictionaryKind(key_kind, value_kind)

        return (array_parser | dictionary_parser | simple_kind_parser).parse(name)  # type: ignore

    @staticmethod
    def any_prop() -> Property:
        return Property("any", "any")


# Split a variable path into its path parts.
# foo.bla -> [foo, bla]
# foo.`bla.bar` -> [foo, bla.bar]
prop_path_parser = (regex("[^`.]+") | variable_dp_backtick).sep_by(dot_dp)
array_index_re = re.compile(r"\[(\d+|\*)]")


class PropertyPath:
    @staticmethod
    def from_path(path: str) -> PropertyPath:
        # remove index accesses from the path (e.g. [23] -> "[]", [*] -> "[]")
        no_index = array_index_re.sub("[]", path)
        return PropertyPath(prop_path_parser.parse(no_index), no_index)

    def __init__(self, path: Sequence[Optional[str]], str_rep: Optional[str] = None):
        self.path = path
        self.path_str = str_rep if str_rep else ".".join(a if a else "" for a in self.path)

    @property
    def root(self) -> bool:
        return not bool(self.path)

    def child(self, part: Optional[str]) -> PropertyPath:
        update = list(self.path)
        update.append(part)
        return PropertyPath(update)

    @property
    def last_part(self) -> Optional[str]:
        return self.path[-1] if self.path else None

    def same_as(self, other: PropertyPath) -> bool:
        """
        Checks if the given path is the same this path.
        Note: the path may include "holes" marked as None.
              The holes mean positions that have to be ignored.
              [A,B,C] same_as [A,None,C] same_as [A,None,None] same_as [None,None,None]
        """
        if len(other.path) == len(self.path):
            for left, right in zip(self.path, other.path):
                if left is not None and right is not None and left != right:
                    return False
            return True
        else:
            return False

    def __repr__(self) -> str:
        return self.path_str

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, PropertyPath) and other.path_str == self.path_str

    def __hash__(self) -> int:
        return hash(self.path_str)


EmptyPath = PropertyPath([], "")


@define(order=True, hash=True, frozen=True)
class ResolvedProperty:
    # The path of the resolved property in a complex kind
    path: PropertyPath
    # The metadata of the property (name etc.)
    prop: Property
    # the resolved kind of this property
    kind: Kind

    @property
    def simple_kind(self) -> SimpleKind:
        return self.kind if isinstance(self.kind, SimpleKind) else AnyKind()


class Kind(ABC):
    def __init__(self, fqn: str):
        self.fqn = fqn

    def __eq__(self, other: object) -> bool:
        return self.__dict__ == other.__dict__ if isinstance(other, Kind) else False

    def coerce(self, value: JsonElement, **kwargs: bool) -> JsonElement:
        coerced = self.coerce_if_required(value, **kwargs)
        return coerced or value

    @abstractmethod
    def coerce_if_required(self, value: JsonElement, **kwargs: bool) -> Optional[JsonElement]:
        """
        Coerces the given value to this kind.
        - if obj is already correct: return None
        - if obj can not be coerced: return None
        - if obj can be coerced: return the value
        """

    @abstractmethod
    def check_valid(self, obj: JsonElement, **kwargs: bool) -> ValidationResult:
        """
        Validate given object against definition:
        - if obj is valid: return None
        - if obj is not valid and can be coerced: return the coerced value.
        - if obj is not valid and can not be coerced: raise an exception
        """

    def resolve(self, model: Dict[str, Kind]) -> None:
        pass

    def kind_hierarchy(self) -> Set[str]:
        return {self.fqn}

    def package(self) -> Optional[str]:
        return self.fqn.rsplit(".", 1)[0] if "." in self.fqn else None

    # noinspection PyUnusedLocal
    @staticmethod
    def from_json(js: Json, _: type = object, **kwargs: object) -> Kind:
        if "inner" in js:
            inner = Kind.from_json(js["inner"])
            return ArrayKind(inner)
        elif "fqn" in js and "runtime_kind" in js and js["runtime_kind"] in simple_kind_to_type:
            fqn = js["fqn"]
            rk = js["runtime_kind"]
            if "source_fqn" in js and "converter" in js and "reverse_order" in js:
                return TransformKind(fqn, rk, js["source_fqn"], js["converter"], js["reverse_order"])
            elif rk == "string":
                minimum = js.get("min_length")
                maximum = js.get("max_length")
                p = js.get("pattern")
                e = js.get("enum")
                return StringKind(fqn, minimum, maximum, p, e)
            elif rk in ("int32", "int64", "float", "double"):
                minimum = js.get("minimum")
                maximum = js.get("maximum")
                e = js.get("enum")
                return NumberKind(fqn, rk, minimum, maximum, e)
            elif rk == "datetime":
                return DateTimeKind(fqn)
            elif rk == "date":
                return DateKind(fqn)
            elif rk == "duration":
                return DurationKind(fqn)
            elif rk == "boolean":
                return BooleanKind(fqn)
            else:
                raise TypeError(f"Unhandled runtime kind: {rk}")
        elif js.get("fqn") == "any" and js.get("runtime_kind") == "any":
            return AnyKind()
        elif "fqn" in js and ("properties" in js or "bases" in js):
            props = list(map(lambda p: from_js(p, Property), js.get("properties", [])))
            bases: Optional[List[str]] = js.get("bases")
            allow_unknown_props = js.get("allow_unknown_props", False)
            successor_kinds = js.get("successor_kinds")
            aggregate_root = js.get("aggregate_root", True)
            return ComplexKind(
                js["fqn"], bases if bases else [], props, allow_unknown_props, successor_kinds, aggregate_root
            )
        else:
            raise JSONDecodeError("Given type can not be read.", json.dumps(js), 0)

    def sort_json(self, js: Json) -> Json:
        """
        Sort the given json element according to the kind definition.
        A model which define properties in a specific order, the json object also uses the same order.
        This will walk the complete json structure.
        """
        if isinstance(js, dict) and isinstance(self, ComplexKind):
            prop_order = {rp.name: idx for idx, rp in enumerate(self.all_props())}
            sub = {
                k: self.property_kind_of(k, AnyKind()).sort_json(v) if isinstance(v, dict) else v for k, v in js.items()
            }
            # a property which is unknown is moved to the end with no specific order
            return dict(sorted(sub.items(), key=lambda k: prop_order.get(k[0], sys.maxsize)))
        elif isinstance(js, dict):
            # noinspection PyTypeChecker
            return dict(sorted(js.items()))
        else:
            return js

    def nested_complex_kinds(self) -> List[ComplexKind]:
        if isinstance(self, ComplexKind):
            return [self]
        elif isinstance(self, ArrayKind):
            return self.inner.nested_complex_kinds()
        elif isinstance(self, DictionaryKind):
            return self.key_kind.nested_complex_kinds() + self.value_kind.nested_complex_kinds()
        else:
            return []


simple_kind_to_type: Dict[str, Type[Union[str, int, float, bool]]] = {
    "string": str,
    "int32": int,
    "int64": int,
    "float": float,
    "double": float,
    "boolean": bool,
    "date": str,
    "datetime": str,
    "duration": str,
}


class SimpleKind(Kind, ABC):
    def __init__(self, fqn: str, runtime_kind: str, reverse_order: bool = False):
        super().__init__(fqn)
        self.runtime_kind = runtime_kind
        self.reverse_order = reverse_order

    def coerce_if_required(self, value: JsonElement, **kwargs: bool) -> Optional[JsonElement]:
        """
        Take a user defined value and transform it into a machine queryable value.
        Example:
            - "10s" as string -> "10s"
            - "10s" as boolean -> false
            - "10s" as duration -> "10s"
            - "10s" as date -> now + 10 seconds (depending on local time)
        :param value: the value from the user
        :return: the coerced value from the system
        """
        return None

    def as_json(self) -> Json:
        return {"fqn": self.fqn, "runtime_kind": self.runtime_kind}

    # noinspection PyUnusedLocal
    @staticmethod
    def to_json(obj: SimpleKind, **kw_args: object) -> Json:
        return obj.as_json()


class AnyKind(SimpleKind):
    def __init__(self) -> None:
        super().__init__("any", "any")

    def __new__(cls) -> AnyKind:
        if cls.__singleton is None:
            cls.__singleton = super(AnyKind, cls).__new__(cls)
        return cls.__singleton

    def check_valid(self, obj: JsonElement, **kwargs: bool) -> ValidationResult:
        return None

    __singleton: Optional[AnyKind] = None


class StringKind(SimpleKind):
    def __init__(
        self,
        fqn: str,
        min_length: Optional[int] = None,
        max_length: Optional[int] = None,
        pattern: Optional[str] = None,
        enum: Optional[Set[str]] = None,
    ):
        super().__init__(fqn, "string")
        self.min_length = min_length
        self.max_length = max_length
        self.pattern = pattern
        self.pattern_compiled = if_set(pattern, re.compile)
        self.enum = enum
        self.valid_fn = validate_fn(
            check_type_fn(str, "string"),
            check_fn(
                self.pattern_compiled,
                lambda p, obj: p.fullmatch(obj) is not None,
                f"does not conform to regex: {self.pattern}",
            ),
            check_fn(self.enum, lambda x, obj: obj in x, f"should be one of: {self.enum}"),
            check_fn(self.min_length, lambda x, obj: len(obj) >= x, f"does not have minimal length: {self.min_length}"),
            check_fn(self.max_length, lambda x, obj: len(obj) <= x, f"is too long! Allowed: {self.max_length}"),
        )

    def check_valid(self, obj: JsonElement, **kwargs: bool) -> ValidationResult:
        if obj is None:
            return None
        elif isinstance(obj, str):
            return self.valid_fn(obj)
        coerced = self.coerce_if_required(obj, **kwargs)
        if coerced is not None:
            self.valid_fn(coerced)
            return coerced
        else:
            raise AttributeError(f"Expected type {self.runtime_kind} but got {type(obj).__name__}")

    def coerce_if_required(self, value: JsonElement, **kwargs: bool) -> Optional[str]:
        if value is None:
            return value
        if isinstance(value, str):
            return None
        else:
            return str(value)

    def as_json(self) -> Json:
        js = super().as_json()
        if self.pattern:
            js["pattern"] = self.pattern
        if self.enum:
            js["enum"] = self.enum
        if self.min_length:
            js["min_length"] = self.min_length
        if self.max_length:
            js["max_length"] = self.max_length
        return js


class NumberKind(SimpleKind):
    def __init__(
        self,
        fqn: str,
        runtime_kind: str,
        minimum: Union[None, float, int] = None,
        maximum: Union[None, float, int] = None,
        enum: Optional[Set[Union[float, int]]] = None,
    ):
        super().__init__(fqn, runtime_kind)
        self.minimum = minimum
        self.maximum = maximum
        self.enum = enum
        self.valid_fn = validate_fn(
            check_type_fn(int, "int") if runtime_kind == "int" else self.check_float,
            check_fn(self.enum, lambda x, obj: obj in x, f"should be one of: {self.enum}"),
            check_fn(self.minimum, lambda x, obj: obj >= x, f"should be greater or equals than: {self.minimum}"),
            check_fn(self.maximum, lambda x, obj: obj <= x, f"should be smaller or equals than: {self.maximum}"),
        )

    @staticmethod
    def check_float(obj: Any) -> ValidationResult:
        if isinstance(obj, (int, float)):
            return None
        else:
            raise AttributeError(f"Expected number but got {obj}")

    def check_valid(self, obj: JsonElement, **kwargs: bool) -> ValidationResult:
        if obj is None:
            return None
        elif isinstance(obj, (int, float)):
            return self.valid_fn(obj)
        coerced = self.coerce_if_required(obj, **kwargs)
        if coerced is not None:
            self.valid_fn(coerced)
            return coerced
        else:
            raise AttributeError(f"Expected type {self.runtime_kind} but got {type(obj).__name__}")

    def coerce_if_required(self, value: JsonElement, **kwargs: bool) -> Optional[Union[int, float]]:
        if value is None:
            return value
        elif isinstance(value, (int, float)):
            return None
        else:
            try:
                return int(value) if self.runtime_kind in ("int32", "int64") else float(value)  # type: ignore
            except ValueError:
                return None

    def as_json(self) -> Json:
        js = super().as_json()
        if self.enum:
            js["enum"] = self.enum
        if self.minimum:
            js["minimum"] = self.minimum
        if self.maximum:
            js["maximum"] = self.maximum
        return js


class BooleanKind(SimpleKind):
    def __init__(self, fqn: str):
        super().__init__(fqn, "boolean")
        self.valid_fn = check_type_fn(bool, "boolean")

    def check_valid(self, obj: JsonElement, **kwargs: bool) -> ValidationResult:
        if obj is True or obj is False or obj is None:
            return None
        coerced = self.coerce_if_required(obj, **kwargs)
        if coerced is not None:
            return coerced
        else:
            raise AttributeError(f"Expected type boolean but got {type(obj).__name__}")

    def coerce_if_required(self, value: JsonElement, **kwargs: bool) -> Optional[bool]:
        if value is None:
            return value
        elif isinstance(value, bool):
            return None
        elif isinstance(value, str) and value.lower() in ("true", "yes", "on"):
            return True
        elif isinstance(value, str) and value.lower() in ("false", "no", "off"):
            return False
        else:
            return None


class DurationKind(SimpleKind):
    def __init__(self, fqn: str):
        super().__init__(fqn, "duration")
        self.valid_fn = validate_fn(check_type_fn(str, "duration"), self.check_duration)

    def check_duration(self, v: Any) -> None:
        if not DurationRe.fullmatch(v):
            raise AttributeError(f"Wrong format for duration: {v}. Examples: 1yr, 3mo, 3d4h3min1s, 3days and 2hours")

    def check_valid(self, obj: JsonElement, **kwargs: bool) -> ValidationResult:
        return self.valid_fn(obj)

    def coerce_if_required(self, value: Any, **kwargs: bool) -> Optional[str]:
        try:
            return f"{int(duration_parser.parse(value))}s" if kwargs.get("normalize", True) else None
        except Exception:
            return None


class DateTimeKind(SimpleKind):
    Format = "%Y-%m-%dT%H:%M:%SZ"
    DateTimeRe = re.compile("\\d{4}-\\d{2}-\\d{2}T\\d{2}:\\d{2}:\\d{2}Z")

    def __init__(self, fqn: str):
        super().__init__(fqn, "datetime")
        self.valid_fn = validate_fn(check_type_fn(str, "datetime"), self.check_datetime)

    @staticmethod
    def parse_datetime(date_string: str) -> Optional[datetime]:
        try:
            return datetime.strptime(date_string, DateTimeKind.Format)
        except ValueError:
            return None

    @staticmethod
    def check_datetime(obj: Any) -> ValidationResult:
        def parse_datetime() -> str:
            parsed = datetime.fromisoformat(str(obj))
            utc_parsed = datetime.fromtimestamp(parsed.timestamp(), tz=timezone.utc)
            return utc_parsed.strftime(DateTimeKind.Format)

        return None if DateTimeKind.DateTimeRe.fullmatch(obj) else parse_datetime()

    def check_valid(self, obj: JsonElement, **kwargs: bool) -> ValidationResult:
        return self.valid_fn(obj)

    def coerce_if_required(self, value: JsonElement, **kwargs: bool) -> Optional[str]:
        try:
            if value is None:
                return value
            elif not isinstance(value, str):
                return None
            elif self.DateTimeRe.fullmatch(value):
                return None
            elif DurationRe.fullmatch(value):
                return self.from_duration(value)
            else:
                return self.from_datetime(value)
        except Exception:
            return None

    @staticmethod
    def from_datetime(value: str) -> str:
        try:
            dt = datetime.fromisoformat(value)
        except Exception:
            dt = parse(value)
        if (
            not dt.tzinfo
            or dt.tzinfo.utcoffset(None) is None
            or dt.tzinfo.utcoffset(None).total_seconds() != 0  # type: ignore
        ):
            dt = dt.astimezone(timezone.utc)
        return dt.strftime(DateTimeKind.Format)

    @staticmethod
    def from_duration(value: str, now: datetime = utc()) -> str:
        # in case of duration, compute the timestamp as: now + duration
        delta = duration(value)
        instant = now + delta
        return instant.strftime(DateTimeKind.Format)


class DateKind(SimpleKind):
    Format = "%Y-%m-%d"
    DateRe = re.compile("\\d{4}-\\d{2}-\\d{2}")

    def __init__(self, fqn: str):
        super().__init__(fqn, "date")
        self.valid_fn = validate_fn(check_type_fn(str, "date"), self.check_date)

    @staticmethod
    def check_date(obj: Any) -> ValidationResult:
        return None if DateKind.DateRe.fullmatch(obj) else date.fromisoformat(obj)

    def check_valid(self, obj: JsonElement, **kwargs: bool) -> ValidationResult:
        return self.valid_fn(obj)

    def coerce_if_required(self, value: JsonElement, **kwargs: bool) -> Optional[str]:
        try:
            if value is None:
                return value
            elif not isinstance(value, str):
                return None
            elif DurationRe.fullmatch(value):
                # in case of duration, compute the timestamp as: today + duration
                delta = duration(value)
                at = date.today() + delta
                return at.isoformat()
            else:
                return parse(value).date().strftime(DateKind.Format)
        except Exception:
            return None


class TransformKind(SimpleKind):
    """
    Transform kinds can be used to derive attributes in a complex kind from other attributes.
    It is important, that the transformed attribute does not exist in the original complex kind!
    It is a SimpleKind, since it can be queried directly as if it would be available as part of the json.

    A transformed kind takes a source value of kind source_kind and transforms it using a function
    into the destination kind.

    :param fqn: the fully qualified name of this kind.
    :param source_fqn: the underlying runtime kind.
    :param destination_fqn: the destination kind that is used in the data store.
    :param converter: name of converter. See transform_kind_convert.py for a dict of possible converter names.
    """

    def __init__(self, fqn: str, source_fqn: str, destination_fqn: str, converter: str, reverse_order: bool):
        # note: source_fqn and runtime_kind are considered the same.
        # the synthetic property does not introduce a new type, but translates types.
        super().__init__(fqn, source_fqn, reverse_order)
        self.destination_fqn: str = destination_fqn
        self.source_kind: Optional[SimpleKind] = None
        self.destination_kind: Optional[SimpleKind] = None
        self.converter = converter
        self.source_to_destination, self.destination_to_source = converters[converter]

    def coerce_if_required(self, value: JsonElement, **kwargs: bool) -> Optional[JsonElement]:
        if value is None:
            return None
        elif self.source_kind:
            coerced_source = self.source_kind.coerce(value)
            real = self.source_to_destination(coerced_source)  # type: ignore
            return real
        else:
            return None

    def transform(self, value: Any) -> Any:
        if value is None:
            return None
        elif self.destination_kind:
            real_coerced = self.destination_kind.coerce(value)
            synth = self.destination_to_source(real_coerced)  # type: ignore
            return synth
        else:
            raise AttributeError(f"Synthetic kind is not resolved: {self.fqn}")

    def check_valid(self, obj: JsonElement, **kwargs: bool) -> ValidationResult:
        # this function is called during import for supplied values.
        # synthetic values are never supplied
        raise AttributeError(f"TransformKind {self.fqn} is not allowed to be supplied.")

    def resolve(self, model: Dict[str, Kind]) -> None:
        source = model.get(self.runtime_kind)
        destination = model.get(self.destination_fqn)
        if source and destination and isinstance(source, SimpleKind) and isinstance(destination, SimpleKind):
            source.resolve(model)
            destination.resolve(model)
            self.source_kind = source
            self.destination_kind = destination
        else:
            raise AttributeError(f"Underlying kind not known: {self.destination_fqn}")

    def as_json(self) -> Json:
        return {
            "fqn": self.fqn,
            "runtime_kind": self.runtime_kind,
            "source_fqn": self.destination_fqn,
            "converter": self.converter,
            "reverse_order": self.reverse_order,
        }


class ArrayKind(Kind):
    def __init__(self, inner: Kind):
        super().__init__(f"{inner.fqn}[]")
        self.inner = inner

    def resolve(self, model: Dict[str, Kind]) -> None:
        self.inner.resolve(model)

    def check_valid(self, obj: JsonElement, **kwargs: bool) -> ValidationResult:
        if isinstance(obj, dict):
            # list is expected, but object found
            raise AttributeError("Expected property is a json object not an array!")

        coerced = self.coerce_if_required(obj, **kwargs)
        for elem in coerced or obj:  # type: ignore
            self.inner.check_valid(elem, **kwargs)
        return coerced

    def coerce_if_required(self, value: JsonElement, **kwargs: bool) -> Optional[List[JsonElement]]:
        has_coerced = False
        if isinstance(value, dict):
            return None
        elif not isinstance(value, list):
            # in case of simple type, we can make it an array
            value = [value]
            has_coerced = True

        def check(item: Any) -> ValidationResult:
            nonlocal has_coerced
            res = self.inner.coerce_if_required(item, **kwargs)
            if res is None:
                return item
            else:
                has_coerced = True
                return res

        mapped = [check(elem) for elem in value]
        return mapped if has_coerced else None

    @staticmethod
    def mk_array(kind: Kind, depth: int) -> Kind:
        return kind if depth == 0 else ArrayKind(ArrayKind.mk_array(kind, depth - 1))


class DictionaryKind(Kind):
    def __init__(self, key_kind: Kind, value_kind: Kind):
        super().__init__(f"dictionary[{key_kind.fqn}, {value_kind.fqn}]")
        self.key_kind = key_kind
        self.value_kind = value_kind

    def check_valid(self, obj: JsonElement, **kwargs: bool) -> ValidationResult:
        if isinstance(obj, dict):
            coerced = self.coerce_if_required(obj, **kwargs)
            for prop, value in (coerced or obj).items():
                part = "key"
                try:
                    self.key_kind.check_valid(prop)
                    part = "value"
                    self.value_kind.check_valid(value)
                except Exception as at:
                    raise AttributeError(f"{part} of {self.fqn} is not valid: {at}") from at
            return coerced
        else:
            raise AttributeError(f"dictionary requires a json object, but got this: {obj}")

    def coerce_if_required(self, value: JsonElement, **kwargs: bool) -> Optional[Json]:
        if isinstance(value, dict):
            coerced: Json = {}
            has_coerced = False
            for p, v in value.items():
                ck = self.key_kind.coerce_if_required(p, **kwargs)
                cv = self.value_kind.coerce_if_required(v, **kwargs)
                coerced[ck or p] = cv or v  # type: ignore
                if ck is not None or cv is not None:
                    has_coerced = True
            return coerced if has_coerced else None
        else:
            return None

    def resolve(self, model: Dict[str, Kind]) -> None:
        self.key_kind.resolve(model)
        self.value_kind.resolve(model)


class ComplexKind(Kind):
    def __init__(
        self,
        fqn: str,
        bases: List[str],
        properties: List[Property],
        allow_unknown_props: bool = False,
        # EdgeType -> possible list of successor kinds
        successor_kinds: Optional[Dict[EdgeType, List[str]]] = None,
        aggregate_root: bool = True,
    ):
        super().__init__(fqn)
        self.bases = bases
        self.properties = properties
        self.allow_unknown_props = allow_unknown_props
        self.successor_kinds = successor_kinds or {}
        self.aggregate_root = aggregate_root
        self.__prop_by_name = {prop.name: prop for prop in properties}
        self.__resolved = False
        self.__resolved_kinds: Dict[str, Tuple[Property, Kind]] = {}
        self.__all_props: List[Property] = list(self.properties)
        self.__resolved_hierarchy: Set[str] = {fqn}
        self.__property_by_path: List[ResolvedProperty] = []
        self.__synthetic_props: List[ResolvedProperty] = []

    def resolve(self, model: Dict[str, Kind]) -> None:
        if not self.__resolved:
            self.__resolved = True
            # resolve properties
            for prop in self.properties:
                kind = prop.resolve(model)
                kind.resolve(model)
                self.__resolved_kinds[prop.name] = (prop, kind)

            # property path -> kind
            self.__property_by_path = ComplexKind.resolve_properties(self)

            # make sure all successor kinds can be resolved
            for names in self.successor_kinds.values():
                for name in names or []:
                    if name not in model:
                        raise AttributeError(f"{name} is not a known kind")

            # resolve the hierarchy
            if not self.is_root():
                for base_name in self.bases:
                    base: Kind = model[base_name]
                    base.resolve(model)
                    if isinstance(base, ComplexKind):
                        self.__resolved_kinds.update(base.__resolved_kinds)
                        self.__all_props = base.__all_props + self.__all_props
                        self.__prop_by_name = {prop.name: prop for prop in self.__all_props}
                        self.__resolved_hierarchy.update(base.__resolved_hierarchy)
                        self.__property_by_path.extend(base.__property_by_path)
            self.__synthetic_props = [p for p in self.__property_by_path if p.prop.synthetic]

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, ComplexKind):
            return (
                self.fqn == other.fqn
                and self.properties == other.properties
                and self.bases == other.bases
                and self.allow_unknown_props == other.allow_unknown_props
                and self.successor_kinds == other.successor_kinds
            )
        else:
            return False

    def __contains__(self, name: str) -> bool:
        return name in self.__prop_by_name

    def __getitem__(self, name: str) -> Property:
        return self.__prop_by_name[name]

    def property_kind_of(self, name: str, or_else: Kind) -> Kind:
        maybe = self.__resolved_kinds.get(name)
        return maybe[1] if maybe else or_else

    def property_with_kind_of(self, name: str) -> Optional[Tuple[Property, Kind]]:
        return self.__resolved_kinds.get(name)

    def property_with_kinds(self) -> Iterable[Tuple[Property, Kind]]:
        return self.__resolved_kinds.values()

    def is_root(self) -> bool:
        return not self.bases or (len(self.bases) == 1 and self.bases[0] == self.fqn)

    def kind_hierarchy(self) -> Set[str]:
        return self.__resolved_hierarchy

    def resolved_properties(self) -> List[ResolvedProperty]:
        return self.__property_by_path

    def all_props(self) -> List[Property]:
        return self.__all_props

    def synthetic_props(self) -> List[ResolvedProperty]:
        return self.__synthetic_props

    def check_valid(self, obj: JsonElement, **kwargs: bool) -> ValidationResult:
        if isinstance(obj, dict):
            coerced = self.coerce_if_required(obj, **kwargs)
            for name, value in (coerced or obj).items():
                known = self.__resolved_kinds.get(name, None)
                if known:
                    prop, kind = known
                    if prop.synthetic:
                        # synthetic properties are computed and will not be maintained. Ignore.
                        pass
                    elif value is None:
                        if prop.required:
                            raise AttributeError(f"Required property {prop.name} is undefined!")
                    else:
                        try:
                            kind.check_valid(value, **kwargs)
                        except AttributeError as at:
                            raise AttributeError(
                                f"Kind:{self.fqn} Property:{name} is not valid: {at}: {json.dumps(obj)}"
                            ) from at
                elif name == "kind":
                    pass
                elif not self.allow_unknown_props:
                    raise AttributeError(f"Kind:{self.fqn} Property:{name} is not defined in model!")
            if not kwargs.get("ignore_missing"):
                for prop in self.__all_props:
                    if prop.required and prop.name not in obj:
                        raise AttributeError(
                            f"Kind:{self.fqn} Property:{prop.name} is required and missing in {json.dumps(obj)}"
                        )
            return coerced
        else:
            raise AttributeError("Kind:{self.fqn} expected a complex type but got this: {obj}")

    def coerce_if_required(self, value: JsonElement, **kwargs: bool) -> Optional[Json]:
        if isinstance(value, dict):
            result: Json = {}
            has_coerced = False
            for n, v in value.items():
                known = self.__resolved_kinds.get(n, None)
                if known:
                    prop, kind = known
                    if prop.synthetic:
                        # synthetic properties are computed and will not be maintained. Ignore.
                        pass
                    elif v is None:
                        result[n] = None
                    else:
                        coerced = kind.coerce_if_required(v, **kwargs)
                        has_coerced |= coerced is not None
                        result[n] = coerced if coerced is not None else v
                else:
                    result[n] = v
            return result if has_coerced else None
        else:
            return None

    def create_yaml(self, elem: JsonElement, initial_level: int = 0) -> str:
        def safe_string(s: str, default_style: Optional[str] = None) -> str:
            return remove_suffix(
                yaml.dump(s, allow_unicode=True, width=sys.maxsize, default_style=default_style), "\n...\n"
            ).strip()

        def walk_element(e: JsonElement, kind: Kind, indent: int, cr_on_object: bool = True) -> str:
            if isinstance(e, dict):
                result = "\n" if cr_on_object else ""
                prepend = "  " * indent
                for prop, value in e.items():
                    description = None
                    sub: Kind = AnyKind()
                    if isinstance(kind, ComplexKind):
                        maybe_prop = kind.property_with_kind_of(prop)
                        if maybe_prop:
                            description = maybe_prop[0].description
                            sub = maybe_prop[1]
                    elif isinstance(kind, DictionaryKind):
                        sub = kind.value_kind
                    str_value = walk_element(value, sub, indent + 1)
                    if description:
                        for line in description.splitlines():
                            result += f"{prepend}# {line}\n"
                    maybe_space = "" if str_value.startswith("\n") else " "
                    safe_prop = safe_string(prop)
                    result += f"{prepend}{safe_prop}:{maybe_space}{str_value}\n"
                return result.rstrip()
            elif isinstance(e, list) and e:
                prepend = "  " * indent + "-"
                sub = kind.inner if isinstance(kind, ArrayKind) else kind
                result = "\n"
                for item in e:
                    item_str = walk_element(item, sub, indent + 1, False).lstrip()
                    result += f"{prepend} {item_str}\n"
                return result.rstrip()
            elif isinstance(e, list):
                return "[]"
            elif isinstance(e, str):
                return safe_string(e, "'")
            elif e is None:
                return "null"
            elif e is True:
                return "true"
            elif e is False:
                return "false"
            else:
                return str(e)

        return walk_element(elem, self, initial_level)

    @staticmethod
    def resolve_properties(
        complex_kind: ComplexKind,
        from_path: PropertyPath = EmptyPath,
        maybe_visited: Optional[Dict[str, PropertyPath]] = None,
    ) -> List[ResolvedProperty]:
        visited = maybe_visited or {}
        result: List[ResolvedProperty] = []

        def path_for(
            prop: Property, kind: Kind, path: PropertyPath, array: bool = False, add_prop_to_path: bool = True
        ) -> None:
            prop_name = f"{prop.name}[]" if array else prop.name
            # Detect object cycles: remember the path when we have visited this property.
            # More complex cycles can be detected that way - leave it simple for now.
            key = f"{prop_name}:{prop.kind}"
            if key in visited and prop_name in visited[key].path:
                return
            visited[key] = path
            relative = path.child(prop_name) if add_prop_to_path else path
            if isinstance(kind, SimpleKind):
                result.append(ResolvedProperty(relative, prop, kind))
            elif isinstance(kind, ArrayKind):
                if name := relative.last_part:
                    result.append(ResolvedProperty(relative, Property(name, kind.fqn), kind))
                path_for(prop, kind.inner, path, True)
            elif isinstance(kind, DictionaryKind):
                child = relative.child(None)
                if name := relative.last_part:
                    result.append(ResolvedProperty(relative, Property(name, kind.fqn), kind))
                    # Any child path accessing this dictionary will get a property of value kind.
                    value = kind.value_kind
                    result.append(ResolvedProperty(child, Property("any", value.fqn), value))
                path_for(prop, kind.value_kind, child, add_prop_to_path=False)
            elif isinstance(kind, ComplexKind):
                if name := relative.last_part:
                    result.append(ResolvedProperty(relative, Property(name, kind.fqn), kind))
                result.extend(ComplexKind.resolve_properties(kind, relative, visited))

        for x in complex_kind.properties:
            path_for(x, complex_kind.__resolved_kinds[x.name][1], from_path)

        return result


string_kind = StringKind("string")
int32_kind = NumberKind("int32", "int32")
int64_kind = NumberKind("int64", "int64")
float_kind = NumberKind("float", "float")
double_kind = NumberKind("double", "double")
any_kind = AnyKind()
boolean_kind = BooleanKind("boolean")
date_kind = DateKind("date")
datetime_kind = DateTimeKind("datetime")
duration_kind = DurationKind("duration")

predefined_kinds = [
    string_kind,
    int32_kind,
    int64_kind,
    float_kind,
    double_kind,
    any_kind,
    boolean_kind,
    date_kind,
    datetime_kind,
    duration_kind,
    TransformKind("trafo.duration_to_datetime", "duration", "datetime", "duration_to_datetime", reverse_order=True),
    ComplexKind(
        "graph_root",
        [],
        [
            Property("name", "string", False, None, "The name of this node."),
            Property("tags", "dictionary[string, string]", False, None, "All attached tags of this node."),
        ],
        allow_unknown_props=True,
    ),
    ComplexKind(
        "predefined_properties",
        [],
        [
            Property("kind", "string", False, None, "The kind property of every node."),
            Property("ctime", "datetime", False, None, "datetime when the node has been created."),
            Property("age", "trafo.duration_to_datetime", False, SyntheticProperty(["ctime"])),
            Property("last_update", "trafo.duration_to_datetime", False, SyntheticProperty(["mtime"])),
            Property("last_access", "trafo.duration_to_datetime", False, SyntheticProperty(["atime"])),
            Property("expires", "datetime", False, None, "datetime when the node expires."),
        ],
    ),
]
predefined_kinds_by_name = {k.fqn: k for k in predefined_kinds}

allowed_simple_type_changes: List[Tuple[Optional[str], Optional[str]]] = [
    ("string", "duration"),
    ("string", "date"),
    ("duration", "string"),
    ("date", "string"),
    ("any", None),
    (None, "any"),
]


class Model:
    @staticmethod
    def empty() -> Model:
        return Model({})

    @staticmethod
    def from_kinds(kinds: List[Kind]) -> Model:
        all_kinds = kinds + predefined_kinds
        kind_dict = {kind.fqn: kind for kind in all_kinds}
        for kind in all_kinds:
            kind.resolve(kind_dict)
        return Model(kind_dict)

    def __init__(self, kinds: Dict[str, Kind]):
        self.kinds = kinds
        self.__property_kind_by_path: List[ResolvedProperty] = list(
            # several complex kinds might have the same property
            # reduce the list by hash over the path.
            {
                r.path: r
                for c in kinds.values()
                if isinstance(c, ComplexKind) and c.aggregate_root
                for r in c.resolved_properties()
            }.values()
        )

    def __contains__(self, name_or_object: Union[str, Json]) -> bool:
        if isinstance(name_or_object, str):
            return name_or_object in self.kinds
        elif isinstance(name_or_object, dict) and "kind" in name_or_object:
            return name_or_object["kind"] is self.kinds
        else:
            return False

    def __getitem__(self, name_or_object: Union[str, Json]) -> Kind:
        if isinstance(name_or_object, str):
            return self.kinds[name_or_object]
        elif isinstance(name_or_object, dict) and "kind" in name_or_object:
            return self.kinds[name_or_object["kind"]]
        else:
            raise KeyError(f"Expected string or json with a 'kind' property as key but got: {name_or_object}")

    def __len__(self) -> int:
        return len(self.kinds)

    def get(self, name_or_object: Union[str, Json]) -> Optional[Kind]:
        if isinstance(name_or_object, str):
            return self.kinds.get(name_or_object)
        elif isinstance(name_or_object, dict) and "kind" in name_or_object:
            return self.kinds.get(name_or_object["kind"])
        else:
            return None

    def complex_kinds(self) -> List[ComplexKind]:
        return [k for k in self.kinds.values() if isinstance(k, ComplexKind)]

    def property_by_path(self, path_: str) -> ResolvedProperty:
        path = PropertyPath.from_path(path_)
        found: Optional[ResolvedProperty] = first(lambda prop: prop.path.same_as(path), self.__property_kind_by_path)
        # if the path is not known according to known model: it could be anything.
        return found if found else ResolvedProperty(path, Property.any_prop(), AnyKind())

    def kind_by_path(self, path_: str) -> Kind:
        return self.property_by_path(path_).kind

    def coerce(self, js: Json) -> Json:
        try:
            kind: Kind = self[js["kind"]]
            return kind.coerce(js)  # type: ignore
        except KeyError as ex:
            raise AttributeError(
                f'No kind definition found for {js["kind"]}' if "kind" in js else f"No attribute kind found in {js}"
            ) from ex

    def check_valid(self, js: Json, **kwargs: bool) -> ValidationResult:
        try:
            kind: Kind = self[js["kind"]]
            return kind.check_valid(js, **kwargs)
        except KeyError as ex:
            raise AttributeError(
                f'No kind definition found for {js["kind"]}' if "kind" in js else f"No attribute kind found in {js}"
            ) from ex

    def graph(self) -> MultiDiGraph:
        graph = MultiDiGraph()

        def handle_complex(cx: ComplexKind) -> None:
            # do not handle the same complex kind more than once
            if cx.fqn in graph and "data" in graph.nodes[cx.fqn]:
                return

            graph.add_node(cx.fqn, data=cx)
            # inheritance
            if not cx.is_root():
                for base in cx.bases:
                    graph.add_edge(cx.fqn, base, f"inheritance_{cx.fqn}_{base}", type="inheritance")

            # properties
            for _, prop_kind in cx.property_with_kinds():
                for cpl in prop_kind.nested_complex_kinds():
                    graph.add_edge(cx.fqn, cpl.fqn, f"property_{cx.fqn}_{cpl.fqn}", type="property")
                    handle_complex(cpl)

            # dependency
            for name, successors in cx.successor_kinds.items():
                for successor in successors or []:
                    graph.add_edge(
                        cx.fqn, successor, f"successor_{cx.fqn}_{successor}_{name}", type="successor", edge_type=name
                    )

        for kind in self.kinds.values():
            if isinstance(kind, ComplexKind):
                handle_complex(kind)
            elif isinstance(kind, ArrayKind) and isinstance(kind.inner, ComplexKind):
                handle_complex(kind.inner)
            elif isinstance(kind, DictionaryKind) and isinstance(kind.value_kind, ComplexKind):
                handle_complex(kind.value_kind)

        return graph

    def update_kinds(self, kinds: List[Kind]) -> Model:

        # Create a list of kinds that have changed to the existing model
        to_update = []
        for elem in kinds:
            existing_props = self.kinds.get(elem.fqn)
            if elem.fqn not in predefined_kinds_by_name and (not existing_props or existing_props != elem):
                to_update.append(elem)

        # Short circuit, if there are no changes
        if not to_update:
            return self

        def update_is_valid(from_kind: Kind, to_kind: Kind) -> None:
            def hint() -> str:
                return f"Update {from_kind.fqn}"

            # Allowed changes: The update
            # - does not change it's type (e.g. going from SimpleKind to ComplexKind)
            if type(from_kind) != type(to_kind):  # pylint: disable=unidiomatic-typecheck
                raise AttributeError(f"{hint()} changes an existing property type {from_kind.fqn}")

        # resolve and build dict
        updates = {elem.fqn: elem for elem in to_update}
        updated = {**self.kinds, **updates}
        for elem in to_update:
            elem.resolve(updated)

        # check if updating existing kinds is allowed
        for name in self.kinds.keys() & updates.keys():
            update_is_valid(self.kinds[name], updates[name])

        # check if no property path is overlapping
        def check_no_overlap() -> None:
            existing_complex = [c for c in self.kinds.values() if isinstance(c, ComplexKind) and c.aggregate_root]
            update_complex = [c for c in to_update if isinstance(c, ComplexKind) and c.aggregate_root]
            ex = {p.path: p for k in existing_complex for p in k.resolved_properties()}
            up = {p.path: p for k in update_complex for p in k.resolved_properties()}

            def simple_kind_incompatible(p: PropertyPath) -> bool:
                left = ex[p].simple_kind
                right = up[p].simple_kind
                # consider valid as long as the runtime kind stays the same
                return (left.runtime_kind != right.runtime_kind) and not (
                    any(
                        True
                        for al, ar in allowed_simple_type_changes
                        if (al is None or left.fqn == al) and (ar is None or right.fqn == ar)
                    )
                )

            # Filter out duplicates that have the same kind or any side is any
            non_unique = [a for a in ex.keys() & up.keys() if simple_kind_incompatible(a)]
            if non_unique:
                # PropertyPath -> str
                name_by_kind = {p.path: k.fqn for k in update_complex for p in k.resolved_properties()}
                message = ", ".join(f"{name_by_kind[a]}.{a} ({ex[a].kind.fqn} -> {up[a].kind.fqn})" for a in non_unique)
                raise AttributeError(
                    f"Update not possible: following properties would be non unique having "
                    f"the same path but different type: {message}"
                )

        check_no_overlap()
        return Model(updated)


# register serializer for this class
set_deserializer(Kind.from_json, Kind)
set_serializer(SimpleKind.to_json, SimpleKind)



import logging
from abc import ABC, abstractmethod
import attrs

from typing import AsyncGenerator, Generic, TypeVar, Optional, Type, Callable, List

from arango import DocumentUpdateError, DocumentRevisionError
from jsons import JsonsError

from resotocore.analytics import AnalyticsEventSender
from resotocore.db.async_arangodb import AsyncArangoDB
from resotocore.error import OptimisticLockingFailed
from resotocore.model.typed_model import from_js, type_fqn, to_js
from resotocore.types import Json

log = logging.getLogger(__name__)

T = TypeVar("T")
K = TypeVar("K", bound=str)


class EntityDb(ABC, Generic[K, T]):
    @abstractmethod
    def keys(self) -> AsyncGenerator[K, None]:
        pass

    @abstractmethod
    def all(self) -> AsyncGenerator[T, None]:
        pass

    @abstractmethod
    async def update_many(self, elements: List[T]) -> None:
        pass

    @abstractmethod
    async def get(self, key: K) -> Optional[T]:
        pass

    @abstractmethod
    async def update(self, t: T) -> T:
        pass

    @abstractmethod
    async def delete(self, key: K) -> None:
        pass

    @abstractmethod
    async def delete_value(self, value: T) -> None:
        pass

    @abstractmethod
    async def create_update_schema(self) -> None:
        pass

    @abstractmethod
    async def wipe(self) -> bool:
        pass


class ArangoEntityDb(EntityDb[K, T], ABC):
    def __init__(self, db: AsyncArangoDB, collection_name: str, t_type: Type[T], key_fn: Callable[[T], K]):
        self.db = db
        self.collection_name = collection_name
        self.t_type = t_type
        self.key_of = key_fn

    async def keys(self) -> AsyncGenerator[K, None]:
        with await self.db.keys(self.collection_name) as cursor:
            for element in cursor:
                yield element

    async def all(self) -> AsyncGenerator[T, None]:
        with await self.db.all(self.collection_name) as cursor:
            for element in cursor:
                try:
                    yield from_js(element, self.t_type)
                except JsonsError:
                    log.warning(f"Not able to parse {element} into {type_fqn(self.t_type)}. Ignore.")

    async def update_many(self, elements: List[T]) -> None:
        await self.db.insert_many(self.collection_name, [self.to_doc(e) for e in elements], overwrite=True)

    async def get(self, key: K) -> Optional[T]:
        result = await self.db.get(self.collection_name, key)
        return from_js(result, self.t_type) if result else None

    async def update(self, t: T) -> T:
        doc = self.to_doc(t)
        try:
            result = await self.db.update(self.collection_name, doc, merge=False)
        except DocumentRevisionError as ex:
            raise OptimisticLockingFailed(self.key_of(t)) from ex
        except DocumentUpdateError as ex:
            if ex.error_code == 1202:  # document not found
                result = await self.db.insert(self.collection_name, doc)
            else:
                raise ex
        if hasattr(t, "revision") and "_rev" in result:
            if attrs.has(type(t)):
                t = attrs.evolve(t, revision=result["_rev"])
            else:
                setattr(t, "revision", result["_rev"])
        return t

    async def delete(self, key: K) -> None:
        await self.db.delete(self.collection_name, key, ignore_missing=True)

    async def delete_value(self, value: T) -> None:
        key = self.key_of(value)
        await self.db.delete(self.collection_name, key, ignore_missing=True)

    async def create_update_schema(self) -> None:
        name = self.collection_name
        db = self.db
        if not await db.has_collection(name):
            await db.create_collection(name)

    async def wipe(self) -> bool:
        return await self.db.truncate(self.collection_name)

    def to_doc(self, elem: T) -> Json:
        js = to_js(elem)
        js["_key"] = self.key_of(elem)
        return js


class EventEntityDb(EntityDb[K, T]):
    def __init__(self, db: EntityDb[K, T], event_sender: AnalyticsEventSender, entity_name: str):
        self.db = db
        self.event_sender = event_sender
        self.entity_name = entity_name

    def keys(self) -> AsyncGenerator[K, None]:
        return self.db.keys()

    def all(self) -> AsyncGenerator[T, None]:
        return self.db.all()

    async def update_many(self, elements: List[T]) -> None:
        result = await self.db.update_many(elements)
        await self.event_sender.core_event(f"{self.entity_name}-updated-many", count=len(elements))
        return result

    async def get(self, key: K) -> Optional[T]:
        return await self.db.get(key)

    async def update(self, t: T) -> T:
        result = await self.db.update(t)
        await self.event_sender.core_event(f"{self.entity_name}-updated")
        return result

    async def delete(self, key: K) -> None:
        await self.db.delete(key)
        await self.event_sender.core_event(f"{self.entity_name}-deleted")

    async def delete_value(self, value: T) -> None:
        await self.db.delete_value(value)
        await self.event_sender.core_event(f"{self.entity_name}-deleted")

    async def create_update_schema(self) -> None:
        return await self.db.create_update_schema()

    async def wipe(self) -> bool:
        return await self.db.wipe()



import logging
from typing import Optional, MutableMapping, Union, Tuple

from arango import HTTPClient, Response
from arango.typings import Headers
from requests import Session
from requests.adapters import HTTPAdapter, Retry
from requests_toolbelt import MultipartEncoder

log = logging.getLogger(__name__)


class ArangoHTTPClient(HTTPClient):  # type: ignore
    def __init__(self, timeout: int, verify: Union[str, bool, None]):
        log.info(f"Create ArangoHTTPClient with timeout={timeout} and verify={verify}")
        self.timeout = timeout
        self.verify = verify

    def create_session(self, host: str) -> Session:
        retry_strategy = Retry(
            total=3,
            backoff_factor=1,
            status_forcelist=[429, 500, 502, 503, 504],
            allowed_methods=["HEAD", "GET", "OPTIONS"],
        )
        http_adapter = HTTPAdapter(max_retries=retry_strategy)
        session = Session()
        session.mount("https://", http_adapter)
        session.mount("http://", http_adapter)
        session.verify = self.verify
        return session

    def send_request(
        self,
        session: Session,
        method: str,
        url: str,
        headers: Optional[Headers] = None,
        params: Optional[MutableMapping[str, str]] = None,
        data: Union[str, MultipartEncoder, None] = None,
        auth: Optional[Tuple[str, str]] = None,
    ) -> Response:
        response = session.request(method, url, params, data, headers, auth=auth, timeout=self.timeout)
        return Response(method, response.url, response.headers, response.status_code, response.reason, response.text)



from typing import Any, Optional, Callable, Dict, Sequence, Union, Mapping, Literal

JsonElement = Union[str, int, float, bool, None, Mapping[str, Any], Sequence[Any]]
JsonArray = Sequence[JsonElement]
# See discussion here: https://github.com/python/typing/issues/182
Json = Dict[str, Any]


ValidationResult = Optional[Any]
ValidationFn = Callable[[Any], ValidationResult]

EdgeType = Literal["default", "delete"]



from __future__ import annotations

import asyncio
import logging
from asyncio import Queue, Future
from collections import defaultdict
from contextlib import asynccontextmanager
from attrs import define, field
from datetime import timedelta, datetime
from functools import reduce
from typing import Any, Optional, AsyncGenerator, Dict, List

from resotocore.types import Json
from resotocore.util import utc, Periodic, set_future_result
from resotocore.ids import TaskId, WorkerId

log = logging.getLogger(__name__)


class WorkerTaskName:
    tag = "tag"
    validate_config = "validate_config"


@define(eq=False, frozen=True)
class WorkerTask:
    id: TaskId  # the unique id of the task
    name: str  # the well known name of the task to perform: the worker attaches to this name
    attrs: Dict[str, str]  # all worker attributes need to match those attrs (but the task can define more)
    data: Json
    callback: Future[Json]  # the callers callback. Notify the caller once the task is dones
    timeout: timedelta  # timeout of this task to be performed

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, WorkerTask) and other.id == self.id

    def __hash__(self) -> int:
        return hash(self.id)

    def to_json(self) -> Json:
        return {"task_id": self.id, "task_name": self.name, "data": self.data, "attrs": self.attrs}


@define(order=True, hash=True)
class WorkerTaskResult:
    task_id: TaskId
    result: str
    data: Optional[Json] = None
    error: Optional[str] = None


@define(order=True, hash=True)
class WorkerTaskInProgress:
    task: WorkerTask
    worker: WorkerTaskSubscription
    retry_counter: int
    deadline: datetime


@define(order=True, hash=True)
class WorkerTaskOnHold:
    task: WorkerTask
    retry_counter: int
    deadline: datetime


@define(order=True, hash=True, frozen=True)
class WorkerTaskDescription:
    name: str
    filter: Dict[str, List[str]] = field(factory=dict)


@define(order=True, hash=True, frozen=True)
class WorkerTaskSubscription:
    worker_id: WorkerId
    task: WorkerTaskDescription
    queue: Queue[WorkerTask]

    def __len__(self) -> int:
        return self.queue.qsize()


class WorkerTaskQueue:
    """
    This class implements a simple task queue.
    """

    def __init__(self) -> None:
        # key is the task_name, value is the list of worker subscriptions
        self.worker_by_task_name: Dict[str, List[WorkerTaskSubscription]] = defaultdict(list)
        self.work_count: Dict[str, int] = defaultdict(lambda: 0)
        self.outstanding_tasks: Dict[str, WorkerTaskInProgress] = {}
        self.unassigned_tasks: Dict[str, WorkerTaskOnHold] = {}
        self.lock = asyncio.Lock()  # note: this lock is not reentrant!
        self.outdated_checker: Periodic = Periodic(
            "check_outdated_tasks", self.check_outdated_unassigned_tasks, timedelta(seconds=5)
        )

    async def start(self) -> None:
        await self.outdated_checker.start()

    async def stop(self) -> None:
        await self.outdated_checker.stop()

    @asynccontextmanager
    async def attach(
        self, worker_id: WorkerId, task_descriptions: List[WorkerTaskDescription], queue_size: int = 0
    ) -> AsyncGenerator[Queue[WorkerTask], None]:
        queue: Queue[WorkerTask] = Queue(queue_size)
        subscriptions = [WorkerTaskSubscription(worker_id, td, queue) for td in task_descriptions]

        if len(task_descriptions) == 0:
            raise AttributeError("Need at least one task description to attach!")
        try:
            async with self.lock:
                self.work_count[worker_id] = 0
                for subscription in subscriptions:
                    self.worker_by_task_name[subscription.task.name].append(subscription)
            log.info(f"Worker {worker_id} added to following task queues: {task_descriptions}")
            yield queue
        finally:
            log.info(f"Remove worker: {worker_id}")
            async with self.lock:
                # remove all subscriptions
                for subscription in subscriptions:
                    self.worker_by_task_name[subscription.task.name].remove(subscription)
                # remove counter
                self.work_count.pop(worker_id, None)
                # reschedule open tasks
                open_tasks = [task for task in self.outstanding_tasks.values() if task.worker.worker_id == worker_id]
                await self.__retry_tasks(open_tasks)

    async def add_task(self, task: WorkerTask, retry_count: int = 3) -> None:
        async with self.lock:
            await self.__add_task(task, retry_count)

    async def acknowledge_task(self, worker_id: WorkerId, task_id: TaskId, result: Optional[Json] = None) -> None:
        async with self.lock:
            await self.__acknowledge_task(worker_id, task_id, result)

    async def error_task(self, worker_id: WorkerId, task_id: TaskId, message: str) -> None:
        async with self.lock:
            await self.__error_task(worker_id, task_id, message)

    async def check_outdated_unassigned_tasks(self) -> None:
        now = utc()
        outstanding = [ip for ip in self.outstanding_tasks.values() if ip.deadline < now]
        not_started_outdated = [ns for ns in self.unassigned_tasks.values() if ns.deadline < now]
        async with self.lock:
            await self.__retry_tasks(outstanding)
            for ns in not_started_outdated:
                log.info(f"No worker for task: {ns.task.id}. Give up.")
                set_future_result(ns.task.callback, Exception(f"No worker for task: {ns.task.name}"))
                self.unassigned_tasks.pop(ns.task.id, None)
            # unassigned_task now only holds valid tasks
            for ns in list(self.unassigned_tasks.values()):
                if await self.__add_task(ns.task, ns.retry_counter):
                    self.unassigned_tasks.pop(ns.task.id, None)

    async def __add_task(self, task: WorkerTask, retry_count: int) -> bool:
        def outstanding_tasks(subscription: WorkerTaskSubscription) -> int:
            return self.work_count[subscription.worker_id]

        def can_perform(subscription: WorkerTaskSubscription) -> bool:
            # the filter criteria in the subscription needs to be matched by the task attributes
            # note: the task can define more attributes, that would be ignored
            def matches_task_filter(name: str, filter_list: List[str]) -> bool:
                value = task.attrs.get(name)
                return value in filter_list if value else False

            return all(matches_task_filter(n, f) for n, f in subscription.task.filter.items())

        # all workers that match the task filter
        matching_subscriptions = [wt for wt in self.worker_by_task_name[task.name] if can_perform(wt)]
        if matching_subscriptions:
            # filter the list for worker with the most specific filter (==most task filter)
            max_filter_len = reduce(lambda res, x: max(res, len(x.task.filter)), matching_subscriptions, 0)
            same_filter_len = (a for a in matching_subscriptions if len(a.task.filter) == max_filter_len)
            # sort and use the one with the least amount of outstanding tasks
            subscriptions = sorted(same_filter_len, key=outstanding_tasks)
            sub = subscriptions[0]  # this is the worker with the least amount of work
            # todo: store task in db
            self.outstanding_tasks[task.id] = WorkerTaskInProgress(task, sub, retry_count, utc() + task.timeout)
            await sub.queue.put(task)
            self.work_count[sub.worker_id] = self.work_count[sub.worker_id] + 1
            return True
        else:
            self.outstanding_tasks.pop(task.id, None)
            if task.id not in self.unassigned_tasks:
                self.unassigned_tasks[task.id] = WorkerTaskOnHold(task, retry_count, utc() + task.timeout)
            return False

    async def __acknowledge_task(self, worker_id: WorkerId, task_id: TaskId, result: Optional[Json]) -> None:
        # remove task from internal list
        in_progress = self.outstanding_tasks.get(task_id, None)
        if in_progress:
            if in_progress.worker.worker_id == worker_id:
                self.outstanding_tasks.pop(task_id, None)
                self.work_count[worker_id] = self.work_count[worker_id] - 1
                set_future_result(in_progress.task.callback, result)
                # todo: remove task from database
            else:
                log.info(f"Got result for task {task_id} from wrong worker {worker_id}. outdated?")

    async def __error_task(self, worker_id: WorkerId, task_id: TaskId, message: str) -> None:
        log.warning(f"Task {task_id} yielded an error: {message}")
        in_progress = self.outstanding_tasks.get(task_id, None)
        if in_progress:
            if in_progress.worker.worker_id == worker_id:
                self.outstanding_tasks.pop(task_id, None)
                self.work_count[worker_id] = self.work_count[worker_id] - 1
                set_future_result(in_progress.task.callback, AttributeError(f"Error executing task: {message}"))
                # todo: remove task from database
            else:
                log.info(f"Got error for task {task_id} from wrong worker {worker_id}. outdated?")

    async def __retry_tasks(self, tasks: List[WorkerTaskInProgress]) -> None:
        for task in tasks:
            if task.retry_counter > 0:
                # todo: maybe it still in the queue of the worker?
                # reschedule
                await self.__add_task(task.task, task.retry_counter - 1)
            else:
                log.warning(f"Too many retried executing task {task.task.id}. Give up.")
                self.outstanding_tasks.pop(task.task.id, None)
                self.work_count[task.worker.worker_id] = self.work_count[task.worker.worker_id] - 1
                set_future_result(task.task.callback, TimeoutError("Could not finish the task."))



import functools
from pydoc import locate
from typing import Type, Any

import jsons

from resotocore.types import JsonElement, Json
from resotocore.util import AnyT


@functools.lru_cache(maxsize=1024)
def class_by_name(class_name: str) -> type:
    clazz = locate(class_name)
    return clazz if clazz is not None else dict  # type: ignore


def class_fqn(obj: Any) -> str:
    return type_fqn(obj.__class__)


def type_fqn(tpe: type) -> str:
    module = tpe.__module__
    return tpe.__name__ if module is None or module == str.__class__.__module__ else module + "." + tpe.__name__


def from_js(json: JsonElement, clazz: Type[AnyT]) -> AnyT:
    return jsons.load(json, cls=clazz) if clazz != dict else json  # type: ignore


def to_js(node: Any, **kwargs: Any) -> Json:
    """
    Use this method, if the given node is known as complex object,
    so the result will be a json object.
    Otherwise: use to_json directly.
    """
    return to_json(node, **kwargs)  # type: ignore


def to_json(node: Any, **kwargs: Any) -> JsonElement:
    # shortcut: assume a dict is already a json value
    if isinstance(node, dict) and not kwargs.get("force_dict", False):
        return node
    return jsons.dump(  # type: ignore
        node,
        strip_privates=True,
        strip_microseconds=True,
        # strip_class_variables=True,
        **kwargs,
    )


def to_js_str(node: Any) -> str:
    return jsons.dumps(node, strip_privates=True)  # type: ignore



from __future__ import annotations

from abc import ABC, abstractmethod
from attrs import define
from typing import Optional, AsyncIterator, List

from jsons import set_deserializer, set_serializer

from resotocore.model.model import Model, Kind
from resotocore.types import Json
from resotocore.ids import ConfigId


@define(order=True, hash=True, frozen=True)
class ConfigEntity:
    id: ConfigId
    config: Json
    revision: Optional[str] = None

    # noinspection PyUnusedLocal
    @staticmethod
    def from_json(js: Json, _: type = object, **kwargs: object) -> ConfigEntity:
        if "id" in js and "config" in js:
            return ConfigEntity(js["id"], js["config"], js.get("_rev"))
        else:
            raise AttributeError(f"Can not parse a ConfigEntity from this json: {js}")

    # noinspection PyUnusedLocal
    @staticmethod
    def to_json(o: ConfigEntity, **kw_args: object) -> Json:
        return dict(id=o.id, config=o.config, _rev=o.revision)


@define(order=True, hash=True, frozen=True)
class ConfigValidation:
    id: str
    external_validation: bool = False


class ConfigHandler(ABC):
    @abstractmethod
    def list_config_ids(self) -> AsyncIterator[ConfigId]:
        pass

    @abstractmethod
    async def get_config(self, cfg_id: ConfigId) -> Optional[ConfigEntity]:
        pass

    @abstractmethod
    async def put_config(self, cfg: ConfigEntity, validate: bool = True) -> ConfigEntity:
        pass

    @abstractmethod
    async def patch_config(self, cfg: ConfigEntity) -> ConfigEntity:
        pass

    @abstractmethod
    async def delete_config(self, cfg_id: ConfigId) -> None:
        pass

    @abstractmethod
    async def get_configs_model(self) -> Model:
        pass

    @abstractmethod
    async def update_configs_model(self, kinds: List[Kind]) -> Model:
        pass

    @abstractmethod
    def list_config_validation_ids(self) -> AsyncIterator[str]:
        pass

    @abstractmethod
    async def get_config_validation(self, cfg_id: str) -> Optional[ConfigValidation]:
        pass

    @abstractmethod
    async def put_config_validation(self, validation: ConfigValidation) -> ConfigValidation:
        pass

    @abstractmethod
    async def config_yaml(self, cfg_id: ConfigId, revision: bool = False) -> Optional[str]:
        pass


# register serializer for this class
set_deserializer(ConfigEntity.from_json, ConfigEntity)
set_serializer(ConfigEntity.to_json, ConfigEntity)



from abc import ABC, abstractmethod
from typing import Optional

from resotocore.query.model import Query


class QueryParser(ABC):
    """
    A query parser is able to parse a simple string into a query.
    """

    @abstractmethod
    async def parse_query(
        self, to_parse: str, on_section: Optional[str], *, omit_section_expansion: bool = False, **env: str
    ) -> Query:
        """
        Parse given string into a query.
        The query might contain expandable sections.
        :param to_parse: the string to parse.
        :param on_section: interpret the query relative to given section.
               If the section is none, all paths are considered absolute.
        :param omit_section_expansion: if this flag is true, the query is not interpreted relative to section.
        :return: the parsed query.
        """



from __future__ import annotations
import asyncio
import logging
from collections import deque
from datetime import timedelta, datetime
from typing import Any, MutableSequence, Optional

from aiohttp import ClientSession
from posthog import Client

from resotocore.analytics import AnalyticsEventSender, AnalyticsEvent
from resotocore.db import SystemData
from resotocore.util import uuid_str, Periodic, utc

log = logging.getLogger(__name__)


class PostHogEventSender(AnalyticsEventSender):
    """
    This analytics event sender uses PostHog (https://posthog.com) to capture all analytics events.
    """

    def __init__(
        self,
        system_data: SystemData,
        flush_at: int = 10000,
        interval: timedelta = timedelta(minutes=1),
        host: str = "https://analytics.some.engineering",
        client_flush_interval: float = 0.5,
        client_retries: int = 3,
    ):
        """
        Create a new PostHog sender.
        :param system_data: information about the current executing system.
        :param flush_at: number of events that can queue up, before the queue is flushed directly.
        :param interval: the frequency when the queue should be flushed.
        :param host: Only here for testing purposes.
        :param client_flush_interval: only here for testing purposes.
        :param client_retries: only here for testing purposes.
        """
        # Note: the client also has the ability to queue events with a flush interval.
        # Sadly: in order to shutdown one has to wait the full interval in worst case!
        # In order to circumvent this behaviour, the queue is maintained here with a configurable interval.
        # In case of shutdown all events are flushed directly and the system is stopped.
        # Note 2: the public api-key is fetched on demand
        self.client = Client(
            api_key="n/a", host=host, flush_interval=client_flush_interval, max_retries=client_retries, gzip=True
        )
        self.run_id = uuid_str()  # create a unique id for this instance run
        self.system_data = system_data
        self.queue: MutableSequence[AnalyticsEvent] = deque()
        self.flush_at = flush_at
        self.flusher = Periodic("flush_analytics", self.flush, interval)
        self.lock = asyncio.Lock()
        self.last_fetched: Optional[datetime] = None
        self.session: Optional[ClientSession] = None

    async def capture(self, event: AnalyticsEvent) -> None:
        """
        Capture a single event by adding it to an internal queue.
        The queue is flushed by a scheduled function.
        Only in the rare case when the queue size reached its maximum the queue will be flushed directly.
        """
        async with self.lock:
            self.queue.append(event)

        if len(self.queue) >= self.flush_at:
            await self.flush()

    async def refresh_public_api_key(self) -> None:
        """
        The API key is public but not static, so we need to refresh it periodically.
        """
        try:
            if not self.session:
                self.session = ClientSession()
            async with self.session.get("https://cdn.some.engineering/posthog/public_api_key") as resp:
                api_key = (await resp.text()).strip()
                self.client.api_key = api_key
                for consumer in self.client.consumers:
                    consumer.api_key = api_key
                self.last_fetched = utc()
        except Exception as ex:
            log.debug(f"Could not fetch latest api key. Will use the current one. {ex}")

    async def flush(self) -> None:
        """
        Flush all events to the posthog server.
        """
        # check, if we need to fetch or refresh the public api key
        if not self.last_fetched:
            await self.refresh_public_api_key()
            sd = self.system_data
            self.client.identify(sd.system_id, {"run_id": self.run_id, "created_at": sd.created_at})
        elif (utc() - self.last_fetched) > timedelta(hours=1):
            await self.refresh_public_api_key()

        # acquire the lock, send all events to the client and clear the queue
        async with self.lock:
            for event in self.queue:
                self.client.capture(
                    distinct_id=self.system_data.system_id,
                    event=event.kind,
                    properties={**event.context, **event.counters, "run_id": self.run_id},  # type: ignore
                    timestamp=event.at,
                )
            self.queue.clear()

    async def __aenter__(self) -> PostHogEventSender:
        return await self.start()

    async def __aexit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        return await self.stop()

    async def start(self) -> PostHogEventSender:
        await self.flusher.start()
        return self

    async def stop(self) -> None:
        await self.flusher.stop()
        await self.flush()
        if self.session:
            await self.session.close()
        self.client.shutdown()
        logging.info("AnalyticsEventSender closed.")



from __future__ import annotations
import asyncio
import json
import logging
from abc import ABC
from asyncio import Task
from contextlib import suppress
from attrs import define
from datetime import timedelta
from multiprocessing import Process, Queue
from queue import Empty
from typing import Optional, Union, AsyncGenerator, Any, Generator, List

from aiostream import stream
from aiostream.core import Stream

from resotocore.analytics import AnalyticsEventSender, InMemoryEventSender, AnalyticsEvent
from resotocore.async_extensions import run_async
from resotocore.core_config import CoreConfig
from resotocore.db.db_access import DbAccess
from resotocore.db.graphdb import GraphDB
from resotocore.db.model import GraphUpdate
from resotocore.db.deferred_edge_db import PendingDeferredEdges
from resotocore.dependencies import db_access, setup_process, reset_process_start_method
from resotocore.error import ImportAborted
from resotocore.model.graph_access import GraphBuilder
from resotocore.model.model import Model
from resotocore.types import Json
from resotocore.ids import TaskId
from resotocore.util import utc, uuid_str, shutdown_process

log = logging.getLogger(__name__)


class ProcessAction(ABC):
    """
    Base class to exchange commands between processes.
    Important: all messages must be serializable in order to get pickled/unpickled.
    """


@define
class ReadElement(ProcessAction):
    """
    Read an incoming element:
    - either a line of text or
    - a complete json element
    Parent -> Child: for every incoming data line.
    """

    elements: List[Union[bytes, Json]]
    task_id: Optional[str]

    def jsons(self) -> Generator[Json, Any, None]:
        return (e if isinstance(e, dict) else json.loads(e) for e in self.elements)


@define
class MergeGraph(ProcessAction):
    """
    Merge the graph that has been read so far.
    Parent -> Child: once EOF of the incoming graph is reached.
    """

    graph: str
    change_id: str
    is_batch: bool = False
    task_id: Optional[TaskId] = None


@define
class EmitAnalyticsEvent(ProcessAction):
    """
    Emit this message to the event sender.
    Child -> Parent: to have the event from the child process propagated to the parent process.
    """

    event: AnalyticsEvent


class PoisonPill(ProcessAction):
    """
    Sentence the process to die.
    Parent -> Child: when the update process is interrupted.
    """


@define
class Result(ProcessAction):
    result: Union[GraphUpdate, str]

    def get_value(self) -> GraphUpdate:
        if isinstance(self.result, str):
            raise ImportAborted(self.result)
        else:
            return self.result


BatchSize = 100000


class DbUpdaterProcess(Process):
    """
    This update class implements Process and is supposed to run as separate process.
    Note: default starting method is supposed to be "spawn".

    This process has 2 queues to read input from and write output to.
    All elements in either queues are of type ProcessAction.

    The parent process should stream the raw commands of graph to this process via ReadElement objects.
    Once the MergeGraph action is received, the graph gets imported.
    From here the parent expects result messages from the child.
    All events happen in the child are forwarded to the parent via EmitEvent.
    Once the graph update is done, a result is send.
    The result is either an exception in case of failure or a graph update in success case.
    """

    def __init__(self, read_queue: Queue[ProcessAction], write_queue: Queue[ProcessAction], config: CoreConfig) -> None:
        super().__init__(name="merge_update")
        self.read_queue = read_queue
        self.write_queue = write_queue
        self.config = config

    def next_action(self) -> ProcessAction:
        try:
            # graph is read into memory. If the sender does not send data in a given amount of time,
            # we raise an exception and abort the update.
            return self.read_queue.get(True, 90)
        except Empty as ex:
            raise ImportAborted("Merge process did not receive any data for more than 90 seconds. Abort.") from ex

    async def merge_graph(self, db: DbAccess) -> GraphUpdate:  # type: ignore
        model = Model.from_kinds([kind async for kind in db.model_db.all()])
        builder = GraphBuilder(model)
        nxt = self.next_action()
        while isinstance(nxt, ReadElement):
            for element in nxt.jsons():
                builder.add_from_json(element)
            log.debug(f"Read {int(BatchSize / 1000)}K elements in process")
            nxt = self.next_action()
        if isinstance(nxt, PoisonPill):
            log.debug("Got poison pill - going to die.")
            shutdown_process(0)
        elif isinstance(nxt, MergeGraph):
            log.debug("Graph read into memory")
            builder.check_complete()
            graphdb = db.get_graph_db(nxt.graph)
            outer_edge_db = db.pending_deferred_edge_db
            _, result = await graphdb.merge_graph(builder.graph, model, nxt.change_id, nxt.is_batch)
            if nxt.task_id and builder.deferred_edges:
                await outer_edge_db.update(PendingDeferredEdges(nxt.task_id, utc(), nxt.graph, builder.deferred_edges))
                log.debug(f"Updated {len(builder.deferred_edges)} pending outer edges for collect task {nxt.task_id}")
            return result

    async def setup_and_merge(self) -> GraphUpdate:
        sender = InMemoryEventSender()
        _, _, sdb = DbAccess.connect(self.config.args, timedelta(seconds=3), verify=self.config.run.verify)
        db = db_access(self.config, sdb, sender)
        result = await self.merge_graph(db)
        for event in sender.events:
            self.write_queue.put(EmitAnalyticsEvent(event))
        return result

    def run(self) -> None:
        try:
            # Entrypoint of the new service
            setup_process(self.config.args, self.config)
            log.info(f"Import process started: {self.pid}")
            result = asyncio.run(self.setup_and_merge())
            self.write_queue.put(Result(result))
            log.info(f"Update process done: {self.pid} Exit.")
            shutdown_process(0)
        except Exception as ex:
            # not all exceptions can be pickled. Use string representation.
            self.write_queue.put(Result(repr(ex)))
            log.error(f"Update process interrupted. Preemptive Exit. {ex}", exc_info=ex)
            shutdown_process(1)


async def merge_graph_process(
    db: GraphDB,
    event_sender: AnalyticsEventSender,
    config: CoreConfig,
    content: AsyncGenerator[Union[bytes, Json], None],
    max_wait: timedelta,
    maybe_batch: Optional[str],
    task_id: Optional[TaskId],
) -> GraphUpdate:
    change_id = maybe_batch if maybe_batch else uuid_str()
    write: Queue[ProcessAction] = Queue()
    read: Queue[ProcessAction] = Queue()
    updater = DbUpdaterProcess(write, read, config)  # the process reads from our write queue and vice versa
    stale = timedelta(seconds=5).total_seconds()  # consider dead communication after this amount of time
    deadline = utc() + max_wait
    dead_adjusted = False

    async def send_to_child(pa: ProcessAction) -> bool:
        alive = updater.is_alive()
        if alive:
            await run_async(write.put, pa, True, stale)
        return alive

    def read_results() -> Task[GraphUpdate]:
        async def read_forever() -> GraphUpdate:
            nonlocal deadline
            nonlocal dead_adjusted
            while utc() < deadline:
                # After exit of updater: adjust the deadline once
                if not updater.is_alive() and not dead_adjusted:
                    log.debug("Import process done or dead. Adjust deadline.")
                    deadline = utc() + timedelta(seconds=30)
                    dead_adjusted = True
                try:
                    action = await run_async(read.get, True, stale)
                    if isinstance(action, EmitAnalyticsEvent):
                        await event_sender.capture(action.event)
                    elif isinstance(action, Result):
                        return action.get_value()
                except Empty:
                    # empty is fine
                    pass
            raise ImportAborted(f"Import process died. (ExitCode: {updater.exitcode})")

        return asyncio.create_task(read_forever())

    task: Optional[Task[GraphUpdate]] = None
    result: Optional[GraphUpdate] = None
    try:
        reset_process_start_method()  # other libraries might have tampered the value in the mean time
        updater.start()
        task = read_results()  # concurrently read result queue
        chunked: Stream = stream.chunks(content, BatchSize)
        async with chunked.stream() as streamer:
            async for lines in streamer:
                if not await send_to_child(ReadElement(lines, task_id)):
                    # in case the child is dead, we should stop
                    break
        await send_to_child(MergeGraph(db.name, change_id, maybe_batch is not None, task_id))
        result = await task  # wait for final result
        return result
    finally:
        if task is not None and not task.done():
            task.cancel()
        if not result:
            # make sure the change is aborted in case of transaction
            log.info(f"Abort update manually: {change_id}")
            await db.abort_update(change_id)
        await send_to_child(PoisonPill())
        await run_async(updater.join, stale)
        if updater.is_alive():
            log.warning(f"Process is still alive after poison pill. Terminate process {updater.pid}")
            with suppress(Exception):
                updater.terminate()
            await asyncio.sleep(3)
        if updater.is_alive():
            log.warning(f"Process is still alive after terminate. Kill process {updater.pid}")
            with suppress(Exception):
                updater.kill()
            await run_async(updater.join)
        if not updater.is_alive():
            with suppress(Exception):
                updater.close()



import asyncio
from datetime import timedelta
from typing import Optional, AsyncIterator, List

import yaml

from resotocore.config import ConfigHandler, ConfigEntity, ConfigValidation
from resotocore.db.configdb import ConfigEntityDb, ConfigValidationEntityDb
from resotocore.db.modeldb import ModelDb
from resotocore.message_bus import MessageBus, CoreMessage
from resotocore.model.model import Model, Kind, ComplexKind
from resotocore.types import Json
from resotocore.util import uuid_str, deep_merge
from resotocore.worker_task_queue import WorkerTaskQueue, WorkerTask, WorkerTaskName
from resotocore.ids import TaskId, ConfigId


class ConfigHandlerService(ConfigHandler):
    def __init__(
        self,
        cfg_db: ConfigEntityDb,
        validation_db: ConfigValidationEntityDb,
        model_db: ModelDb,
        task_queue: WorkerTaskQueue,
        message_bus: MessageBus,
    ) -> None:
        self.cfg_db = cfg_db
        self.validation_db = validation_db
        self.model_db = model_db
        self.task_queue = task_queue
        self.message_bus = message_bus

    async def coerce_and_check_model(self, cfg_id: ConfigId, config: Json, validate: bool = True) -> Json:
        model = await self.get_configs_model()

        final_config = {}
        if validate:
            for key, value in config.items():
                if key in model:
                    try:
                        value_kind = model[key]
                        coerced = value_kind.check_valid(value, normalize=False)
                        final_config[key] = value_kind.sort_json(coerced or value)
                    except Exception as ex:
                        raise AttributeError(f"Error validating section {key}: {ex}") from ex
                else:
                    final_config[key] = value
        else:
            final_config = config

        # If an external entity needs to approve this change.
        # Method throws if config is not valid according to external approval.
        validation = await self.validation_db.get(cfg_id)
        if validation and validation.external_validation and validate:
            await self.acknowledge_config_change(cfg_id, final_config)

        # If we come here, everything is fine
        return final_config

    def list_config_ids(self) -> AsyncIterator[ConfigId]:
        return self.cfg_db.keys()

    async def get_config(self, cfg_id: ConfigId) -> Optional[ConfigEntity]:
        return await self.cfg_db.get(cfg_id)

    async def put_config(self, cfg: ConfigEntity, validate: bool = True) -> ConfigEntity:
        coerced = await self.coerce_and_check_model(cfg.id, cfg.config, validate)
        existing = await self.cfg_db.get(cfg.id)
        if not existing or existing.config != cfg.config:
            result = await self.cfg_db.update(ConfigEntity(cfg.id, coerced, cfg.revision))
            await self.message_bus.emit_event(CoreMessage.ConfigUpdated, dict(id=result.id, revision=result.revision))
            return result
        else:
            return existing

    async def patch_config(self, cfg: ConfigEntity) -> ConfigEntity:
        current = await self.cfg_db.get(cfg.id)
        current_config = current.config if current else {}
        coerced = await self.coerce_and_check_model(cfg.id, deep_merge(current_config, cfg.config))
        result = await self.cfg_db.update(ConfigEntity(cfg.id, coerced, current.revision if current else None))
        await self.message_bus.emit_event(CoreMessage.ConfigUpdated, dict(id=result.id, revision=result.revision))
        return result

    async def delete_config(self, cfg_id: ConfigId) -> None:
        await self.cfg_db.delete(cfg_id)
        await self.validation_db.delete(cfg_id)
        await self.message_bus.emit_event(CoreMessage.ConfigDeleted, dict(id=cfg_id))

    def list_config_validation_ids(self) -> AsyncIterator[str]:
        return self.validation_db.keys()

    async def get_config_validation(self, cfg_id: str) -> Optional[ConfigValidation]:
        return await self.validation_db.get(cfg_id)

    async def put_config_validation(self, validation: ConfigValidation) -> ConfigValidation:
        return await self.validation_db.update(validation)

    async def get_configs_model(self) -> Model:
        kinds = [kind async for kind in self.model_db.all()]
        return Model.from_kinds(list(kinds))

    async def update_configs_model(self, kinds: List[Kind]) -> Model:
        # load existing model
        model = await self.get_configs_model()
        # make sure the update is valid
        updated = model.update_kinds(kinds)
        # store all updated kinds
        await self.model_db.update_many(kinds)
        return updated

    async def config_yaml(self, cfg_id: ConfigId, revision: bool = False) -> Optional[str]:
        config = await self.get_config(cfg_id)
        if config:
            model = await self.get_configs_model()

            yaml_str = ""
            for num, (key, value) in enumerate(config.config.items()):
                maybe_kind = model.get(key)
                if isinstance(maybe_kind, ComplexKind):
                    part = maybe_kind.create_yaml(value, initial_level=1)
                    if num > 0:
                        yaml_str += "\n"
                    yaml_str += key + ":" + part
                else:
                    yaml_str += yaml.dump({key: value}, sort_keys=False, allow_unicode=True)

            # mix the revision into the yaml document
            if revision and config.revision:
                yaml_str += (
                    "\n\n# This property is not part of the configuration but defines the revision "
                    "of this document.\n# Please leave it here to avoid conflicting writes.\n"
                    f'_revision: "{config.revision}"'
                )

            return yaml_str
        else:
            return None

    async def acknowledge_config_change(self, cfg_id: str, config: Json) -> None:
        """
        In case an external entity should acknowledge this config change.
        This method either return, which signals success or throws an exception.
        """
        future = asyncio.get_event_loop().create_future()
        task = WorkerTask(
            TaskId(uuid_str()),
            WorkerTaskName.validate_config,
            {"config_id": cfg_id},
            {"task": WorkerTaskName.validate_config, "config": config},
            future,
            timedelta(seconds=30),
        )
        # add task to queue - do not retry
        await self.task_queue.add_task(task)
        # In case the config is not valid or no worker is available
        # this future will throw an exception.
        # Do not handle it here and let the error bubble up.
        await future



import re
from argparse import ArgumentParser
from functools import lru_cache
from typing import TypeVar, Union, Any, Callable, AsyncIterator, NoReturn, Optional, Awaitable, Tuple, List

from aiostream.core import Stream
from parsy import Parser, regex

from resotolib.parse_util import (
    make_parser,
    literal_dp,
    equals_dp,
    json_value_dp,
    space_dp,
    double_quote_dp,
    double_quoted_string_part_or_esc_dp,
    single_quoted_string_part_or_esc_dp,
    single_quote_dp,
    any_non_white_space_string,
    comma_p,
)
from resotocore.model.graph_access import Section
from resotocore.types import JsonElement
from resotocore.util import AnyT

T = TypeVar("T")
# Allow the function to return either a coroutine or the result directly
Result = Union[T, Awaitable[T]]
JsGen = Union[Stream, AsyncIterator[JsonElement]]
# A sink function takes a stream and creates a result
Sink = Callable[[JsGen], Awaitable[T]]


@make_parser
def key_value_parser() -> Parser:
    key = yield literal_dp
    yield equals_dp
    value = yield json_value_dp
    return key, value


# name=value test=true -> {name: value, test: true}
key_values_parser: Parser = key_value_parser.sep_by(comma_p | space_dp).map(dict)
# anything that is not: | " ' ; \
cmd_token = regex("[^|\"';\\\\]+")
# single and double-quoted string are maintained with quotes: "foo"->"foo", 'foo'->'foo'
# all characters inside the quoted string are not parsed
double_quoted_string = double_quote_dp + double_quoted_string_part_or_esc_dp + double_quote_dp
single_quoted_string = single_quote_dp + single_quoted_string_part_or_esc_dp + single_quote_dp
# same as above, but the surrounding quotes are not preserved: "foo"->foo, 'foo'->foo
double_quoted_raw_string = double_quote_dp >> double_quoted_string_part_or_esc_dp << double_quote_dp
single_quoted_raw_string = single_quote_dp >> single_quoted_string_part_or_esc_dp << single_quote_dp
# parse \| \" \' \; and unescape it \| -> |
escaped_token = regex("\\\\[|\"';]").map(lambda x: x[1])
# a command are tokens until EOF or pipe (all characters will be preserved)
cmd_with_args_parser = (escaped_token | double_quoted_string | single_quoted_string | cmd_token).at_least(1).concat()

# argument parser which will read the argument list while removing single and double quotes
# command line arguments: foo "bla: 'foo = bla' -> [foo, bla, foo = bla]
args_parts_unquoted_parser = (
    escaped_token | double_quoted_raw_string | single_quoted_raw_string | any_non_white_space_string
).sep_by(space_dp)

# argument parser which will read the argument list while removing single quotes
# Example: "--a \"a or b\" --b 'b or c' --c c d" -> ["--a", "\"a or b\"", "--b", "b or c", "--c", "c", "d"]
args_parts_parser = (
    escaped_token | double_quoted_string | single_quoted_raw_string | any_non_white_space_string
).sep_by(space_dp, min=1)


def strip_quotes(string: str) -> str:
    res = string.strip()
    if res:
        first = res[0]
        if first in "'\"":
            res = res[1 : len(res) - 1] if res.startswith(first) and res.endswith(first) else res  # noqa: E203
    return res


# check if a is a json node element
def is_node(a: Any) -> bool:
    return "id" in a and Section.reported in a if isinstance(a, dict) else False


# check if given object is a json edge element
def is_edge(a: Any) -> bool:
    return "from" in a and "to" in a if isinstance(a, dict) else False


class NoExitArgumentParser(ArgumentParser):
    def error(self, message: str) -> NoReturn:
        raise AttributeError(f"Could not parse arguments: {message}")

    def exit(self, status: int = 0, message: Optional[str] = None) -> NoReturn:
        msg = message if message else "unknown"
        raise AttributeError(f"Could not parse arguments: {msg}")


path_array_index_parser = re.compile(r"([^[]+)\[([^]]*)]")


@lru_cache(maxsize=8192)
def parse_path_index(path: str) -> Tuple[str, Union[bool, int, None]]:
    mm = path_array_index_parser.match(path)
    if mm:
        if mm.group(2) in ("*", ""):
            return mm.group(1), True
        elif mm.group(2).isdigit():
            return mm.group(1), int(mm.group(2))
        else:
            raise ValueError(f"Invalid path index: {path}")
    else:
        return path, None


def js_value_get(element: JsonElement, path_or_name: Union[List[str], str], if_none: AnyT) -> AnyT:
    result = js_value_at(element, path_or_name)
    return result if result and isinstance(result, type(if_none)) else if_none


def js_value_at(element: JsonElement, path_or_name: Union[List[str], str]) -> Optional[Any]:
    path = path_or_name if isinstance(path_or_name, list) else path_or_name.split(".")
    at = len(path)

    def at_idx(current: JsonElement, idx: int) -> Optional[Any]:
        if at == idx:
            return current
        prop, index = parse_path_index(path[idx])
        if current is None or not isinstance(current, dict) or prop not in current:
            return None
        else:
            child = current[prop]
            if isinstance(child, list):
                if index is None:
                    return child
                elif index is True:
                    return [at_idx(e, idx + 1) for e in child]
                elif index < len(child):
                    return at_idx(child[index], idx + 1)
                else:
                    return None
            else:
                return at_idx(child, idx + 1)

    return at_idx(element, 0)
